# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

"""Commands definitions for the `app`."""


from azdata.cli.core.logging import get_logger
from azdata.cli.core.util import raise_command_error
from azdata.cli.core.exceptions import (ControllerError, CliError)
from azdata.cli.core.prompt import (prompt, prompt_y_n)
from azdata.cli.commands.app.spec import Spec
from azdata.cli.commands.app.scafold import Scaffold
from azdata.cli.commands.app.util import (zip_files, silent_remove)
from azdata.cli.commands.app.exceptions import (
    AppError,
    InvalidSpecError
)
from humanfriendly.terminal.spinners import AutomaticSpinner
from knack.prompting import NoTTYException

import os
import time

logger = get_logger(__name__)

IS_WINDOWS = os.name == 'nt'


def app_template_list(client, url=None):
    """
    ```sh
    azdata app template list
    azdata app template --url
    ```
    """
    try:
        # -- fetch template from ulr repository location --
        tpls = []
        if not IS_WINDOWS:
            with AutomaticSpinner('Finding application templates',
                                  show_time=True):
                tpls = client.list()
        else:
            tpls = client.list()

        return tpls
    except ValueError as e:
        raise CliError(e)
    except TypeError:
        if url:
            msg = 'Template repository location [--url] "{}" not found.'. \
                format(url)
        else:
            msg = 'Template content is unavailable or not supported.'
        raise CliError(msg)
    except Exception as e:
        raise_command_error(e)


def app_template_pull(client, name=None, url=None, destination='./templates'):
    """
    ```sh
    azdata app template pull
    azdata app template pull --name --url
    ```
    """

    try:
        # -- fetch template based on runtime --
        if os.path.isdir(os.path.join(destination, name)
                         if name else destination):
            raise FileExistsError()

        # -- download template based on runtime --
        if not IS_WINDOWS:
            with AutomaticSpinner('Fetching application templates',
                                  show_time=True):
                tpl_dir = client.download(name=name, dest=destination)
        else:
            tpl_dir = client.download(name=name, dest=destination)

        stdout = client.stdout
        count = '1 template' if name else '{} templates'.\
            format(len(os.listdir(tpl_dir)))
        stdout('{} downloaded.'.format(count))
        time.sleep(.5)  # stage effect
        stdout('Template(s) created in directory: {}'.
               format(destination))
    except ValueError as e:
        raise CliError(e)
    except TypeError:
        msg = 'Template repository location [--url] "{}" not found.'.format(url)
        raise CliError(msg)
    except FileNotFoundError:
        if url:
            msg = 'Template repository location [--url] "{}" not found.'. \
                format(url)
        else:
            msg = 'Template "{}" is unavailable or not supported.'. \
                format(name)
        raise CliError(msg)
    except FileExistsError:
        raise CliError('Directory already exists, please choose a different '
                       '[--destination]: "{}"'.format(destination))
    except Exception as e:
        raise_command_error(e)


def app_init(
        client,
        spec=None,
        name=None,
        version=None,
        template=None,
        destination=None,
        url=None):

    """
    ```sh
    azdata app init
    azdata app init --name --version --template
    azdata app init --spec --name --version --template --url
    ```
    """

    try:
        if not name:
            name = prompt('Name: ')

        if not version:
            version = prompt('Version: ')

        if not template:
            template = prompt('Template: ')

        if not all([name, version, template]):
            raise ValueError('Missing arguments [--name, --version, --template]'
                             ' are required.')

        if not destination:
            destination = os.getcwd()

        destination = os.path.join(destination, Spec.FILE if spec else name)

        if os.path.isfile(destination) or os.path.isdir(destination):
            raise FileExistsError()

        # -- download template based on runtime --
        if not IS_WINDOWS:
            with AutomaticSpinner('Fetching application template',
                                  show_time=True):
                tpl_dir = client.download(name=template)
        else:
            tpl_dir = client.download(name=template)

        stdout = client.stdout  # The controlled print display

        stdout('{} template downloaded.'.format(template))
        time.sleep(.5)  # stage effect

        # -- generate the app based on template --
        scaffold = Scaffold(name, version, template, destination)
        scaffold.generate(tpl_dir, spec_only=spec)
        stdout('Application {} file written.'.
               format(Spec.FILE))

        if not spec:
            stdout('Application created in directory: {}'.
                   format(name))
    except ValueError as e:
        raise CliError(e)
    except NoTTYException:
        raise CliError('Please specify name, version, and runtime in '
                       'non-interactive mode.')
    except FileNotFoundError:
        if url:
            msg = 'Template repository location [--url] "{}" not found.'.\
                format(url)
        else:
            msg = 'Template "{}" is unavailable or not supported.'.\
                format(template)
        raise CliError(msg)
    except FileExistsError:
        raise CliError('Directory already exists, please choose another '
                       '[--name] or different [--destination]: \n"{}"'.
                       format(destination))
    except Exception as e:
        raise_command_error(e)


def app_create(client, spec):
    """
    ```sh
    azdata app create --spec /path/to/spec/dir
    ```
    """
    tmp_files = []
    try:
        if os.path.isfile(spec):
            raise ValueError('Must provide a directory containing a valid '
                             'spec.yaml file using [--spec]')

        assets = [spec]
        spec = Spec(spec)  # validate on creation

        # -- add clean tmp spec.yaml to asset listing --

        # NOTE: We need to assert the proper IO types in the spec.yaml. This
        # will go away when the server implements it. Here we create a tmp spec
        # based on the one provided.
        tmp_spec = spec.tmp_yaml_file()        
        blacklist = [Spec.FILE]

        # -- deploy --
        archive = '.app-archive.zip'
        tmp_files.append(tmp_spec['directory'])
        tmp_files.append(archive)
        msg = 'Creating application `{0}/{1}`'.format(spec.name, spec.version)

        bundle = zip_files(archive, assets, blacklist=blacklist)

        if not IS_WINDOWS:
            with AutomaticSpinner(msg, show_time=True):
                client.apis.controller.create_app(tmp_spec['path'], bundle)
        else:
            client.apis.controller.create_app(tmp_spec['path'], bundle)

        time.sleep(.5)

        client.stdout('Application `{0}/{1}` successfully created.'.
                      format(spec.name, spec.version))
    except (InvalidSpecError,
            AppError,
            ControllerError,
            IOError,
            ValueError,
            FileNotFoundError) as e:
        raise_command_error(e)
    except NoTTYException:
        raise CliError('Please specify [--yes] argument in non-interactive '
                       'mode.')
    except Exception as e:
        raise_command_error(e)
    finally:
        silent_remove(tmp_files)


def app_update(client, spec=None, yes=None):
    """
    ```sh
    azdata app update
    azdata app update --spec /path/to/spec/dir
    ```
    """

    tmp_files = []
    try:
        # if nothing was provided and there is a spec.yaml in the CWD, use that
        if not spec and os.path.isfile(Spec.FILE):
            spec = '.'

            if not yes:
                yes = prompt_y_n('Are you sure?')
                if not yes:
                    return

        if not spec:
            raise ValueError('Must provide a valid spec.yaml file using '
                             '[--spec]')

        assets = [spec]
        spec = Spec(spec)  # validate on creation

        # -- add clean tmp spec.yaml to asset listing --

        # NOTE: We need to assert the proper IO types in the spec.yaml. This
        # will go away when the server implements it. Here we create a tmp spec
        # based on the one provided.
        tmp_spec = spec.tmp_yaml_file()
        blacklist = [Spec.FILE]

        # -- deploy --
        archive = '.app-archive.zip'
        tmp_files.append(tmp_spec['directory'])
        tmp_files.append(archive)
        msg = 'Updating application `{0}/{1}`'.format(spec.name, spec.version)

        if not IS_WINDOWS:
            with AutomaticSpinner(msg, show_time=True):
                client.apis.controller.update_app(tmp_spec['path'], zip_files(archive, assets, blacklist))
        else:
            client.apis.controller.update_app(tmp_spec['path'], zip_files(archive, assets, blacklist))

        time.sleep(.5)

        client.stdout('Application `{0}/{1}` successfully updated.'.
                      format(spec.name, spec.version))
    except (InvalidSpecError, AppError, ControllerError, ValueError) as e:
        raise_command_error(e)
    except NoTTYException:
        raise CliError('Please specify [--yes] argument in non-interactive '
                       'mode.')
    except Exception as e:
        raise_command_error(e)
    finally:
        silent_remove(tmp_files)


def app_run(client, name, version, inputs=None):
    """
    ```sh
    azdata app run --name <name> --version <version> --inputs x=5,y=2
    ```
    """
    try:
        return client.run_app(name, version, inputs)
    except (AppError, ControllerError, Exception) as e:
        raise_command_error(e)


def app_list(client, name=None, version=None):
    """
    ```sh
    azdata app list --name <name> --version <version>
    ```
    """
    try:
        if not version:
            apps = client.apis.controller.list_apps(name)
        else:
            apps = client.apis.controller.list_app(name, version)

        # -- filter response for less verboseness --
        filtered = []
        for app in apps:
            filtered.append({
                'name': app['name'],
                'version': app['version'],
                'app': app['links']['app'],
                'state': app['state']
            })

        return filtered
    except (AppError, ControllerError, Exception) as e:
        raise_command_error(e)


def app_describe(client, spec=None, name=None, version=None):
    """
    ```sh
    azdata app describe --name <name> --version <version>
    ```
    """
    try:
        # if nothing was provided and there is a spec.yaml in the CWD, use that
        if not spec and not name and os.path.isfile(Spec.FILE):
            spec = '.'

        # must either have a spec.yaml [or] name/version
        if not spec and not all([name, version]):
            raise ValueError('Must provide either a valid spec.yaml file using '
                             '[--spec] or values for [--name, --version].')

        if spec:
            spec = Spec(spec)
            name = spec.name
            version = spec.version

        not_found_msg = 'Application `{0}/{1}` not found.'.format(name, version)
        app = client.apis.controller.list_app(name, version)

        return app[0] if app else not_found_msg
    except ValueError as e:
        raise CliError(e)
    except (InvalidSpecError,
            AppError,
            ControllerError,
            FileNotFoundError,
            Exception) as e:
        raise_command_error(e)


def app_delete(client, name, version):
    """
    ```sh
    azdata app delete --name <name> --version <version>
    ```
    """
    try:
        response = client.apis.controller.delete_app(name, version)
        return response
    except (AppError, ControllerError, Exception)as e:
        raise_command_error(e)
