# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

"""Command definitions for `control`."""

import json
import os
import sys
from datetime import datetime

import azdata.cli.core.deploy as util
import yaml
from azdata.cli.commands.arc import data_controller_properties
from azdata.cli.commands.arc.exceptions import ArcError
from azdata.cli.core.exceptions import (KubernetesError)
from azdata.cli.commands.arc.constants import (
    ARC_NAME,
    CONFIG_DIR,
    HELP_DIR,
    CONTROL_CONFIG_FILENAME,
    CONFIG_FILES,
    DATA_CONTROLLER_CRD,
    POSTGRES_11_CRD,
    POSTGRES_12_CRD,
    SQLMI_CRD
)
from azdata.cli.commands.arc.export_util import (
    ExportType,
    logs_export,
    metrics_export,
    logs_upload,
    metrics_upload,
    _get_log_workspace_credentials_from_env,
    EXPORT_DATA_JSON_SCHEMA,
    EXPORT_FILE_DICT_KEY,
    LOGS_CONFIG_FILENAME,
    EXPORT_SANITIZERS
)
from azdata.cli.commands.arc.common_util import (
    validate_dc_create_params,
    write_output_file
)

from azdata.cli.core.constants import (ARC_GROUP, ARC_API_VERSION, DATA_CONTROLLER_PLURAL)
from azdata.cli.core.debug import (copy_debug_logs, take_dump)
from azdata.cli.core.deploy import DeploymentConfigUtil
from azdata.cli.core.exceptions import CliError
from azdata.cli.core.clients.kubernetes_client import (
    K8sApiException)
from azdata.cli.core.models.data_controller_custom_resource import DataControllerCustomResource
from azdata.cli.core.models.custom_resource_definition import CustomResourceDefinition
from azdata.cli.core.prompt import (prompt_for_input, prompt_for_choice, prompt_y_n)
from azdata.cli.core.serialization import Sanitizer
from humanfriendly.terminal.spinners import AutomaticSpinner
from jsonschema import validate
from knack.prompting import NoTTYException
from urllib3.exceptions import NewConnectionError, MaxRetryError

CONNECTION_RETRY_ATTEMPTS = 12
RETRY_INTERVAL = 5

IS_WINDOWS = os.name == 'nt'


def dc_create(
        client,
        namespace,
        name,
        connectivity_mode,
        resource_group,
        location,
        subscription,
        profile_name=None,
        path=None,
        storage_class=None):
    """
    If an argument is not provided, the user will be prompted for the needed
    values NoTTY Scenario: provide a config_profile, profile_name
    :param storage_class: storage class
    :param client:
    :param namespace: cluster namespace
    :param name: data controller name
    :param profile_name: name of an existing built-in config profile
    :param connectivity_mode: cluster to cloud connectivity mode
    :param resource_group: Azure resource group
    :param location: Azure location
    :param subscription: Azure subscription
    :param path: custom config profile
    :param
    :return:
    """
    try:
        stdout = client.stdout

        # -- Check Kubectl Context --
        util.check_and_set_kubectl_context()

        # Validate params
        validate_dc_create_params(name, namespace, subscription, location,
                                  resource_group, connectivity_mode, profile_name, path)

        #  -- User entered an existing configuration type
        if profile_name:
            path = os.path.join(CONFIG_DIR, profile_name)
            if not os.path.isdir(path):
                raise ValueError("Profile name {0} does not exist.".
                                 format(os.path.basename(profile_name)))

        # -- Configuration Directory --
        if not path:
            profiles = DeploymentConfigUtil.get_config_map(CONFIG_DIR)

            # Prompt the user for a choice between configs
            stdout("Please choose a deployment configuration: ")
            stdout("To see more info please exit create and use command "
                   "[arc dc config list -c <config_profile>]")
            choices = DeploymentConfigUtil.get_config_display_names(profiles)
            # Filter out test profiles
            filtered_choices = list(filter(lambda c: "test" not in c, choices))
            result = prompt_for_choice(filtered_choices, choices[8])

            path = os.path.join(CONFIG_DIR, profiles[result.lower()])

        # -- Required Environment Variables --
        if sys.stdin.isatty():
            util.read_environment_variables(ARC_NAME, True)
        else:
            util.check_environment_variables(ARC_NAME)

        # -- Read json into python dictionary --
        config_object = util.read_config(path, CONTROL_CONFIG_FILENAME)

        dc_cr = DataControllerCustomResource(config_object)

        # Namespace required
        dc_cr.namespace = namespace

        # Name required
        dc_cr.name = name

        # Storage class
        if storage_class:
            dc_cr.storage_class_data = storage_class
            dc_cr.storage_class_logs = storage_class

        # Azure properties
        #
        azure_settings = {
            data_controller_properties.CONNECTION_MODE: connectivity_mode,
            data_controller_properties.LOCATION: location,
            data_controller_properties.RESOURCE_GROUP: resource_group,
            data_controller_properties.SUBSCRIPTION: subscription
        }

        if "settings" not in dc_cr.spec:
            dc_cr.spec["settings"] = {}

        if "controller" not in dc_cr.spec["settings"]:
            dc_cr.spec["settings"]["controller"] = {}

        dc_cr.spec["settings"]["azure"] = azure_settings
        dc_cr.spec["settings"]["controller"][data_controller_properties.DISPLAY_NAME] = name

        # -- Get help documentation for missing values --
        help_object = util.read_config(HELP_DIR, CONTROL_CONFIG_FILENAME)
        # -- Check for missing values in the config object --
        util.check_missing(stdout, False, dc_cr.body, help_object, CONTROL_CONFIG_FILENAME)
        # -- Check if dc config is valid
        util.control_config_check(stdout, dc_cr.body)

        # Create the namespace
        client.apis.control_deploy.create_namespace(dc_cr.namespace)

        crd_files = [POSTGRES_11_CRD, POSTGRES_12_CRD, SQLMI_CRD, DATA_CONTROLLER_CRD]

        for crd_file in crd_files:
            # Create the control plane CRD if it doesn't already exist
            with open(crd_file, "r") as stream:
                temp = yaml.safe_load(stream)
                crd = CustomResourceDefinition(temp)
                util.retry(lambda: client.apis.kubernetes.create_custom_resource_definition(crd),
                           retry_count=CONNECTION_RETRY_ATTEMPTS,
                           retry_delay=RETRY_INTERVAL,
                           retry_method="create custom resource definition",
                           retry_on_exceptions=(NewConnectionError, MaxRetryError, K8sApiException))

        # Create cluster role for metricsdc
        client.create_cluster_role_for_monitoring(dc_cr, namespace)

        # -- attempt to create cluster --
        stdout("")
        stdout("Deploying data controller")
        stdout("")
        stdout("NOTE: Data controller creation can take a significant amount of time "
               "depending on")
        stdout("configuration, network speed, and the number of nodes in the "
               "cluster.")
        stdout("")

        client.dc_create(crd, dc_cr)

        stdout("Data controller successfully deployed.")
    except NoTTYException:
        raise CliError("Please specify --profile-name or --path in non-interactive mode.")
    except ValueError as e:
        raise CliError(e)
    except ArcError as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def dc_endpoint_list(client, endpoint_name=None):
    """
    Retrieves the endpoints of the cluster
    :param client:
    :param endpoint_name:
    :return:
    """
    try:
        return client.apis.controller.get_endpoints(endpoint_name)
    except Exception as e:
        raise CliError(e)


def dc_status_show(client, namespace=None):
    """
    Return the status of the data controller custom resource.
    :param client:
    :param namespace:
    :return:
    """
    try:
        # -- Check Kubectl Context --
        util.check_and_set_kubectl_context()

        if not namespace:
            namespace = client.profile.active_context.namespace

        response = client.apis.kubernetes.list_namespaced_custom_object(
            namespace, group=ARC_GROUP, version=ARC_API_VERSION, plural=DATA_CONTROLLER_PLURAL
        )

        dcs = response.get("items")

        if not dcs:
            client.stdout("No data controller exists in namespace `{}`.".format(namespace))
        else:
            cr = DataControllerCustomResource(dcs[0])

            state = cr.state
            if state:
                client.stdout(state.lower().capitalize())
            else:
                client.stderr(
                    "Status unavailable for data controller `{0}` in namespace `{1}`.".format(cr.name, cr.namespace))

    except Exception as e:
        raise CliError(e)


def dc_config_show(client, namespace=None):
    """
    Return the config of the data controller custom resource.
    :param client:
    :param namespace:
    :return:
    """
    try:
        # -- Check Kubectl Context --
        util.check_and_set_kubectl_context()

        if not namespace:
            namespace = client.profile.active_context.namespace

        response = client.apis.kubernetes.list_namespaced_custom_object(
            namespace, group=ARC_GROUP, version=ARC_API_VERSION, plural=DATA_CONTROLLER_PLURAL
        )

        dcs = response.get("items")

        if not dcs:
            client.stdout("No data controller exists in namespace `{}`.".format(namespace))
        else:
            return dcs[0]

    except Exception as e:
        raise CliError(e)


def dc_delete(client, name, namespace, force=None):
    """
    Deletes the data controller - requires kube config and env var
    :param client:
    :param name:
    :param namespace:
    :param force:
    :return:
    """
    try:
        stdout = client.stdout
        stdout("")
        stdout("This operation will delete everything inside of data controller `{}`".format(name))
        stdout("which includes the Kubernetes secrets and services,")
        stdout("SQL managed instances, PostgreSQL Hyperscale server groups containers. "
               "Data stored on persistent volumes will get deleted if")
        stdout("the storage class reclaim policy is set to delete/recycle.")
        stdout("")

        if not force:
            yes = prompt_y_n("Do you want to continue with deleting "
                             "the data controller `{}`?".format(name))
        else:
            yes = force  # yes|no for no-tty

        if yes != "yes" and yes is not True:
            msg = "Data controller not deleted. Exiting..."
            return msg

        # -- Check Kubectl Context --
        util.check_and_set_kubectl_context()

        # -- Check existence of data controller --
        if not client.apis.kubernetes.namespaced_custom_object_exists(
                name,
                namespace,
                group=ARC_GROUP,
                version=ARC_API_VERSION,
                plural=DATA_CONTROLLER_PLURAL):
            raise CliError("Data controller `{}` does not exist in namespace `{}`.".format(name, namespace))

        # -- Check existence of data services --
        crd_files = [POSTGRES_11_CRD, POSTGRES_12_CRD, SQLMI_CRD]

        for crd_file in crd_files:
            # Create the control plane CRD if it doesn't already exist
            with open(crd_file, "r") as stream:
                temp = yaml.safe_load(stream)
                crd = CustomResourceDefinition(temp)
                cr_list = client.apis.kubernetes.list_namespaced_custom_object(namespace, crd=crd)
                if cr_list["items"]:
                    raise ArcError("Instances of `{}` are deployed. Cannot delete data controller `{}`. "
                                   "Please delete these instances before deleting the data controller."
                                   .format(crd.kind, name))

        stdout("Deleting data controller `{}`.".format(name))

        # -- attempt to delete cluster --
        client.dc_delete(namespace, name)

        # Delete the control plane CRD
        with open(DATA_CONTROLLER_CRD, "r") as stream:
            temp = yaml.safe_load(stream)
            crd = CustomResourceDefinition(temp)
            client.apis.kubernetes.delete_custom_resource_definition(crd)

        stdout("Data controller `{}` deleted successfully.".format(name))

    except NoTTYException:
        raise CliError("Please specify `--force` in non-interactive mode.")
    except ArcError as e:
        raise CliError(e)
    except KubernetesError as e:
        raise CliError(e)
    except K8sApiException as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def dc_config_list(client, config_profile=None):
    """
    Lists available configuration file choices.
    :param client:
    :param config_profile:
    :param config_type:
    :param accept_eula:
    :return:
    """
    try:
        configs = DeploymentConfigUtil.config_list(CONFIG_DIR, config_profile)

        # Filter out test profiles
        filtered_configs = list(filter(lambda c: "test" not in c, configs))

        return filtered_configs

    except ValueError as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def dc_config_init(client, target=None, source=None, force=None):
    """
    Initializes a cluster configuration file for the user.
    :param client:
    :param target:
    :param source:
    :param force:
    :return:
    """
    try:
        try:
            if not target:
                target = prompt_for_input("Custom Config Profile Path:", "custom", False, False)
        except NoTTYException:
            # If non-interactive, default to custom directory
            target = "custom"

        stdout = client.stdout

        # Read the available configs by name
        config_map = DeploymentConfigUtil.get_config_map(CONFIG_DIR)

        if source:
            if source not in config_map.keys():
                raise ValueError("Invalid config source, please consult [dc "
                                 "config list] for available sources")
        elif not source:
            choices = DeploymentConfigUtil.get_config_display_names(config_map)

            # Filter out test profiles
            filtered_choices = list(filter(lambda c: "test" not in c, choices))

            # Prompt the user for a choice between configs
            stdout("Please choose a config:")
            source = prompt_for_choice(filtered_choices, choices[8])

        if os.path.isfile(target):
            raise FileExistsError("Please specify a directory path. Target is a file: {0}".format(target))

        result = DeploymentConfigUtil.save_config_profile(target, source, CONFIG_DIR, CONFIG_FILES, config_map, force)

        client.stdout("Created configuration profile in {}".format(result))

    except ValueError as e:
        raise CliError(e)
    except NoTTYException:
        raise CliError("Please specify target and source in non-interactive mode")
    except Exception as e:
        raise CliError(e)


def dc_config_add(client, config_file, json_values):
    """
    Add new key and value to the given config file
    :param client:
    :param config_file:
    :param json_values:
    :return:
    """
    try:
        config_object = DeploymentConfigUtil.config_add(config_file, json_values)
        DeploymentConfigUtil.write_config_file(config_file, config_object)
    except Exception as e:
        raise CliError(e)


def dc_config_replace(client, config_file, json_values):
    """
    Replace the value of a given key in the given config file
    :param client:
    :param config_file:
    :param json_values:
    :return:
    """
    try:
        config_object = DeploymentConfigUtil.config_replace(config_file, json_values)
        DeploymentConfigUtil.write_config_file(config_file, config_object)
    except Exception as e:
        raise CliError(e)


def dc_config_remove(client, config_file, json_path):
    """
    Remove a key from the given config file
    :param client:
    :param config_file:
    :param json_path:
    :return:
    """
    try:
        config_object = DeploymentConfigUtil.config_remove(config_file, json_path)
        DeploymentConfigUtil.write_config_file(config_file, config_object)
    except Exception as e:
        raise CliError(e)


def dc_config_patch(client, config_file, patch_file):
    """
    Patch a given file against the given config file
    :param client:
    :param config_file:
    :param patch_file:
    :return:
    """
    try:
        config_object = DeploymentConfigUtil.config_patch(config_file, patch_file)
        DeploymentConfigUtil.write_config_file(config_file, config_object)
    except Exception as e:
        raise CliError(e)


def dc_debug_copy_logs(client,
                       namespace,
                       container=None,
                       target_folder=None,
                       pod=None,
                       resource_kind=None,
                       resource_name=None,
                       timeout=0,
                       skip_compress=False,
                       exclude_dumps=False):
    """
    Copy Logs commands - requires kube config
    """
    try:
        # -- Check Kubectl Context --
        util.check_and_set_kubectl_context()

        copy_debug_logs(
            namespace, target_folder, pod, container, resource_kind, resource_name, timeout, skip_compress, exclude_dumps)
    except ArcError as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def dc_debug_dump(
        client, namespace, container="controller", target_folder="./output/dump"):
    """
    Trigger dump for given container and copy out the dump file to given
    output folder
    """
    try:
        # -- Check Kubectl Context --
        util.check_and_set_kubectl_context()

        take_dump(namespace, container, target_folder)
    except ArcError as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def dc_export(client, export_type, path, force=None):
    """
    Export metrics, logs or usage to a file.
    """
    stdout = client.stdout
    cluster_name = client.profile.active_context.namespace
    info_msg = 'This option exports {} of all instances in "{}" to the file: "{}".'

    # -- Check Kubectl Context --
    util.check_and_set_kubectl_context()

    try:
        path = _check_prompt_export_output_file(path, force)
        data_controller = client.get_data_controller(cluster_name)

        content = {
            'exportType': export_type,
            'dataController': data_controller,
            'dataTimestamp': datetime.now().isoformat(),
            'instances': [],
            'data': []
        }

        if export_type.lower() not in ExportType.list():
            raise ValueError('{} is not a supported type. '
                             'Please specify one of the following: {}'.format(export_type, ExportType.list()))

        instances = client.list_all_custom_resource_instances(cluster_name)
        content['instances'] = instances

        active_instances = dict.fromkeys(
            map(lambda x: '{}/{}.{}'.format(x['kind'], x['instanceName'], x['instanceNamespace']), instances))

        deleted_instances = client.list_deleted_resource_instances()

        # Ignored instances which were deleted but subsequently recreated, as their will be updated
        content['deletedInstances'] = list(filter(
            lambda x: '{}/{}.{}'.format(x['kind'], x['instanceName'], x['instanceNamespace'])
                      not in active_instances.keys(),
            deleted_instances))

        if export_type.lower() == ExportType.metrics.value:
            stdout(info_msg.format(
                ExportType.metrics.value,
                client.profile.active_context.namespace,
                path))
            metrics_data = metrics_export(client, instances, data_controller)

            if metrics_data:
                content['data'] = metrics_data
                write_output_file(path, content)
                stdout('Metrics are exported to {}'.format(path))
            else:
                stdout('Failed to get metrics.'
                       'Please ensure you connect to the correct cluster and the instances have metrics.')
        elif export_type.lower() == ExportType.logs.value:
            stdout(info_msg.format(
                ExportType.logs.value,
                client.profile.active_context.namespace,
                path))
            file_list, data_timestamp = logs_export(client, path, instances, data_controller)

            if file_list:
                content['data'] = file_list
                content['dataTimestamp'] = data_timestamp
                write_output_file(path, content)
                stdout('Log files list are exported to {}.'.format(path))
            else:
                stdout('No log is exported.')
        elif export_type.lower() == 'usage':
            usages = client.get_usages()
            if usages:
                usages_data = []
                for u in usages:
                    usages_data.append({
                        "usages": u['usages'], 
                        "signature": u['signature']
                    })
                content['data'] = usages_data
                write_output_file(path, content)
                client.stdout('Usage is exported to {}'.format(path))
            else:
                client.stdout('No usage is exported.')

    except NoTTYException:
        raise CliError('Please specify `--force` in non-interactive mode.')
    except Exception as e:
        raise CliError(e)


def dc_upload(client, path):
    """
    Upload data file exported from a data controller to Azure.
    """

    import uuid

    try:
        if not os.path.exists(path):
            raise FileNotFoundError('Cannot find file: "{}". Please provide the correct file name and try again'.format(
                path))

        with open(path, encoding='utf-8') as input_file:
            data = json.load(input_file)
            data = Sanitizer.sanitize_object(data, EXPORT_SANITIZERS)

            validate(data, EXPORT_DATA_JSON_SCHEMA)
    except Exception as e:
        raise CliError(e)

    # Check expected properties
    #
    for expected_key in EXPORT_FILE_DICT_KEY:
        if expected_key not in data:
            raise ValueError('"{}" is not found in the input file "{}".'.format(expected_key, path))

    export_type = data['exportType']

    if not ExportType.has_value(export_type):
        raise ValueError(
            '"{}" is not a supported type. Please check your input file "{}".'.format(export_type, path))

    # Create/Update shadow resource for data controller
    #
    data_controller = data['dataController']
    client.create_dc_azure_resource(data_controller)

    # Delete shadow resources for resource instances deleted from the cluster in k8s
    #
    deleted = dict()

    for instance in data['deletedInstances']:
        instance_key = '{}/{}.{}'.format(instance['kind'], instance['instanceName'], instance['instanceNamespace'])

        if instance_key not in deleted.keys():
            try:
                client.delete_azure_resource(instance, data_controller)
                deleted[instance_key] = True
            except Exception as e:
                client.stdout('Failed to delete Azure resource for "{}" in "{}".'
                              .format(instance['instanceName'], instance['instanceNamespace']))
                client.stderr(e)
                continue

    # Create/Update shadow resources for resource instances still active in the cluster in k8s
    #
    for instance in data['instances']:
        client.create_azure_resource(instance, data_controller)

    # Upload metrics, logs or usage
    #
    if export_type == ExportType.metrics.value:
        metrics_upload(data['data'])
    elif export_type == ExportType.logs.value:
        customer_id, shared_key = _get_log_workspace_credentials_from_env(client)
        client.stdout('Log Analytics workspace: "{}"'.format(customer_id))
        for file in data['data']:
            with open(file, encoding='utf-8') as input_file:
                data = json.load(input_file)
            logs_upload(client, data['data'], data['dataTimestamp'], customer_id, shared_key)
    elif export_type == "usage":
        if data_controller:
            client.stdout('\n')
            client.stdout('Start uploading usage...')
            correlation_vector = str(uuid.uuid4())
            for usage in data['data']:
                client.upload_usages_dps(data_controller, usage, data['dataTimestamp'], correlation_vector)
            client.stdout('usage upload is done.')
        else:
            client.stdout('No usage to upload.')
    else:
        raise ValueError(
            '"{}" is not a supported type. Please check your input file "{}".'.format(export_type, path))


def arc_resource_kind_list(client):
    """
    Returns the list of available arc resource kinds which can be created in the cluster.
    :param client:
    :return:
    """
    try:
        return client.apis.controller.resource_types_get()

    except Exception as e:
        raise CliError(e)


def arc_resource_kind_get(client, kind, dest="template"):
    """
    Returns a package of crd.json and spec-template.json based on the given kind.
    :param client:
    :param kind:
    :param dest:
    :return:
    """
    try:
        if not os.path.isdir(dest):
            os.makedirs(dest, exist_ok=True)

        if not IS_WINDOWS:
            with AutomaticSpinner("Fetching {0} template".format(kind),
                                  show_time=True):
                crd_response = client.apis.controller.resource_crd_get(kind)
                spec_response = client.apis.controller.resource_spec_get(kind)
        else:
            crd_response = client.apis.controller.resource_crd_get(kind)
            spec_response = client.apis.controller.resource_spec_get(kind)

        crd_pretty = json.dumps(crd_response, indent=4)
        spec_pretty = json.dumps(spec_response, indent=4)

        with open(os.path.join(dest, "crd.json"), "w") as output:
            output.write(crd_pretty)

        with open(os.path.join(dest, "spec.json"), "w") as output:
            output.write(spec_pretty)

        client.stdout("{0} template created in directory: {1}".format(kind, dest))

    except Exception as e:
        raise CliError(e)


def _check_prompt_export_output_file(file_path, force):
    """
    Checks if export output file exists, and prompt if necessary.
    :param file_path:
    :param force:
    :return: final output file path
    """
    # Check if file exists
    export_file_exists = True
    overwritten = False

    while export_file_exists and not overwritten:
        export_file_exists = os.path.exists(file_path)
        if not force and export_file_exists:
            try:
                yes = prompt_y_n('File {} exists already. Overwrite file?'.format(file_path))
            except NoTTYException as e:
                raise NoTTYException('{} Please make the file does not exist in non-interactive environment'.format(e))

            overwritten = True if yes else False

            if overwritten:
                os.remove(file_path)
            else:
                file_path = prompt_for_input('Please provider file name with the path: ')
                export_file_exists = True
                overwritten = False

        elif force:
            overwritten = True
            if export_file_exists:
                os.remove(file_path)

    return file_path
