# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

"""Command definitions for `bdc`."""

from azdata.cli.commands.bdc.exceptions import ClusterError
from azdata.cli.commands.bdc.constants import (
    CONFIG_DIR, HELP_DIR, BDC_CONFIG_FILENAME, CONTROL_CONFIG_FILENAME, CONFIG_FILES
)
from azdata.cli.core.debug import (copy_debug_logs, take_dump)
from azdata.cli.core.util import (raise_command_error, eula_prompt, FileUtil)
from azdata.cli.core.deploy import DeploymentConfigUtil
from azdata.cli.core.exceptions import CliError
from azdata.cli.core.configuration import Configuration
from azdata.cli.core.upgrade import update_upgrade_timeouts
from azdata.cli.core.prompt import (prompt_y_n, prompt_for_input, prompt_for_choice)
from azdata.cli.commands.bdc.constants import (
    CONTROL_SERVICE, SQL_SERVER_SERVICE, HDFS_SERVICE, SPARK_SERVICE, GATEWAY_SERVICE, APP_SERVICE
)
from knack.prompting import NoTTYException
from jsonpath_ng.ext import parse

import azdata.cli.core.deploy as util
import os
import json
import sys


def bdc_create(client, name=None, config_profile=None, accept_eula=None, node_label=None, force=None):
    """
    If an argument is not provided, the user will be prompted for the needed
    values NoTTY Scenario: provide a config_profile, accept_eula
    :param client:
    :param name:
    :param config_profile:
    :param accept_eula:
    :param node_label:
    :param force:
    :return:
    """
    try:
        stdout = client.stdout

        # -- Check if custom config profile exists --
        if config_profile and not os.path.isdir(config_profile):
            # Check in the default directory
            config_profile = os.path.join(CONFIG_DIR, config_profile)
            if not os.path.isdir(config_profile):
                raise ValueError('Config profile {0} does not exist.'.
                                 format(os.path.basename(config_profile)))

        # -- enforce EULA --
        client.profile.accept_eula = eula_prompt(client, accept_eula, True, force, True)

        stdout('\nCluster deployment documentation can be viewed at:')
        stdout('https://aka.ms/bdc-deploy\n')

        # -- Configuration Directory --
        if not config_profile and not force:
            profiles = DeploymentConfigUtil.get_config_map(CONFIG_DIR)

            # Prompt the user for a choice between configs
            stdout('Please choose a deployment configuration: ')
            stdout('To see more info please exit create and use command '
                   '[bdc config list -c <config_profile>]')
            choices = DeploymentConfigUtil.get_config_display_names(profiles)
            result = prompt_for_choice(choices, choices[1])

            config_profile = os.path.join(CONFIG_DIR, profiles[result.lower()])
        elif not config_profile and force:
            raise ValueError('Please specify a configuration profile as an argument via '
                             '\'-c <configuration profile>\' when using --force')

        # -- Check Kubectl Context --
        util.check_and_set_kubectl_context()

        # -- Required Environment Variables --
        if not force and sys.stdin.isatty():
            util.read_environment_variables('cluster')
            stdout('')
        else:
            util.check_environment_variables('cluster')

        configs = dict()
        for config_filename in CONFIG_FILES:
            # -- Read json into python dictionary --
            config_object = util.read_config(config_profile, config_filename)
            # -- Get help documentation for missing values --
            help_object = util.read_config(HELP_DIR, config_filename)
            # -- If security is enabled, check for required security environment variables --
            util.security_check(stdout, force, config_object, name)
            # -- Check for missing values in the config object --
            util.check_missing(stdout, force, config_object, help_object, config_filename)
            configs[config_filename] = config_object

        try:
            config_object = configs[BDC_CONFIG_FILENAME]
            config_object['spec']['controlPlane'] = configs[CONTROL_CONFIG_FILENAME]

            # -- Add cluster nodes label to config if included --
            if node_label is not None:
                config_object["spec"]["controlPlane"]["spec"]["nodeLabel"] = node_label
        except Exception as e:
            raise ValueError('Invalid config profile: {0}'.format(e))

        # -- Check if control config is valid
        util.control_config_check(stdout, configs[CONTROL_CONFIG_FILENAME])

        # -- attempt to create cluster --
        stdout('NOTE: Cluster creation can take a significant amount of time '
               'depending on')
        stdout('configuration, network speed, and the number of nodes in the '
               'cluster.')
        stdout('')

        client.create_bdc(name, config_object)
    except NoTTYException:
        raise CliError('Please specify [--accept-eula, --config-profile] in '
                       'non-interactive mode.')
    except ValueError as e:
        raise CliError(e)
    except ClusterError as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def bdc_delete(client, name, force=None):
    """
    Deletes the cluster - requires kube config and env var
    :param client:
    :param name:
    :param force:
    :return:
    """
    try:
        stdout = client.stdout
        stdout('')
        stdout('This operation will delete everything inside of cluster "{}"'.format(name))
        stdout('which includes the SQL Server containers, Kubernetes secrets and services,')
        stdout('and HDFS containers. Data stored on persistent volumes will get deleted if')
        stdout('the storage class reclaim policy is set to delete/recycle.')
        stdout('Namespace "{}" and other objects created outside of Big Data Cluster'.format(name))
        stdout('will not be dropped.')
        stdout('')

        if not force:
            yes = prompt_y_n('Do you want to continue with deleting '
                             'the cluster "{}"?'.format(name))
        else:
            yes = force  # yes|no for no-tty

        if yes != 'yes' and yes is not True:
            msg = 'Cluster not deleted. Exiting...'
            return msg

        # -- Check Kubectl Context --
        util.check_and_set_kubectl_context()

        # -- attempt to delete cluster --
        client.delete_bdc(name, force)
    except NoTTYException:
        raise CliError('Please specify `--force` in non-interactive mode.')
    except ClusterError as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def bdc_upgrade(client, name, target_version, target_repository=None,
                    controller_timeout=None, stability_threshold=None, component_timeout=None):
    """
    Upgrades all docker images in the cluster to images with the provided tag from the
    optionally provided repository.
    :param client:
    :param target_version:
    :param controller_timeout:
    :param stability_threshold:
    :param component_timeout:
    :return:
    """
    try:
        stdout = client.stdout

        # -- attempt to upgrade cluster --
        if target_repository:
            stdout('Upgrading cluster to repository {0}'.
                format(target_repository))
        stdout('Upgrading cluster to version {0}'.
               format(target_version))
        stdout('')
        stdout('NOTE: Cluster upgrade can take a significant amount of time '
               'depending on')
        stdout('configuration, network speed, and the number of nodes in the '
               'cluster.')
        stdout('')

        # -- Check Kubectl Context --
        util.check_and_set_kubectl_context()

        if controller_timeout or component_timeout:
            update_upgrade_timeouts(name, controller_timeout, stability_threshold, component_timeout)

        client.upgrade_bdc(name, target_version, target_repository)
    except ValueError as e:
        raise CliError(e)
    except ClusterError as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def bdc_endpoint(client, endpoint_name=None):
    """
    Retrieves the endpoints of the cluster
    :param client:
    :param endpoint_name:
    :return:
    """
    try:
        return client.apis.controller.get_endpoints(endpoint_name)
    except Exception as e:
        raise CliError(e)


def bdc_status_show(client, resource=None, all=False):
    """
    Retrieves the status of the cluster
    :param client:
    :param resource: Get all the services that contain this resource
    :param all: Get all of the instances for each resource within their respective services
    :return:
    """
    try:
        return client.apis.controller.get_bdc_status(None, resource, all=all)
    except Exception as e:
        raise CliError(e)


def bdc_control_status_show(client, resource=None, all=False):
    """
    Retrieves the status of the control service
    :param client:
    :param resource: Get this particular resource within this service
    :param all: Get all of the instances for each resource within this service
    :return:
    """
    try:
        return client.apis.controller.get_control_status(resource, all=all)
    except Exception as e:
        raise CliError(e)


def bdc_sql_status_show(client, resource=None, all=False):
    """
    Retrieves the status of the sql-server service
    :param client:
    :param resource: Get this particular resource within this service
    :param all: Get all of the instances for each resource within this service
    :return:
    """
    try:
        return client.apis.controller.get_bdc_status(
            SQL_SERVER_SERVICE, resource, all=all)
    except Exception as e:
        raise CliError(e)


def bdc_hdfs_status_show(client, resource=None, all=False):
    """
    Retrieves the status of the hdfs service
    :param client:
    :param resource: Get this particular resource within this service
    :param all: Get all of the instances for each resource within this service
    :return:
    """
    try:
        return client.apis.controller.get_bdc_status(
            HDFS_SERVICE, resource, all=all)
    except Exception as e:
        raise CliError(e)


def bdc_spark_status_show(client, resource=None, all=False):
    """
    Retrieves the status of the spark service
    :param client:
    :param resource: Get this particular resource within this service
    :param all: Get all of the instances for each resource within this service
    :return:
    """
    try:
        return client.apis.controller.get_bdc_status(
            SPARK_SERVICE, resource, all=all)
    except Exception as e:
        raise CliError(e)


def bdc_gateway_status_show(client, resource=None, all=False):
    """
    Retrieves the status of the gateway service
    :param client:
    :param resource: Get this particular resource within this service
    :param all: Get all of the instances for each resource within this service
    :return:
    """
    try:
        return client.apis.controller.get_bdc_status(
            GATEWAY_SERVICE, resource, all=all)
    except Exception as e:
        raise CliError(e)


def bdc_app_status_show(client, resource=None, all=False):
    """
    Retrieves the status of the app service
    :param client:
    :param resource: Get this particular resource within this service
    :param all: Get all of the instances for each resource within this service
    :return:
    """
    try:
        return client.apis.controller.get_bdc_status(
            APP_SERVICE, resource, all=all)
    except Exception as e:
        raise CliError(e)


def bdc_config_show(client, config_file=None, target=None, json_path=None, force=None):
    """
    Shows the cluster's config or the given config to the user.
    Finds the given json path in the config if a path is given.
    Outputs to target file if given.
    :param client:
    :param config_file:
    :param target:
    :param json_path:
    :param force:
    :return:
    """
    try:
        if not config_file:
            if not client.profile.active_context.access_token:
                raise CliError('Please run \'{0} login\' to first authenticate.'.format(Configuration().CLI_NAME))
            config_response = client.apis.controller.get_cluster()
            config_object = json.loads(config_response)
        else:
            config_object = FileUtil.read_json(config_file)

        if json_path:
            json_path = DeploymentConfigUtil.replace_path(json_path, ['/'], '.')
            expr = parse(json_path)
            find_result = [match.value for match in expr.find(config_object)]
            if len(find_result) > 0:
                config_object = find_result[0]
            if not config_object:
                raise ValueError('Path {json_path} resulted in no value for config {config}'
                                 .format(json_path=json_path, config=json.dumps(config_object, indent=4)))

        json_data_pretty = json.dumps(config_object, indent=4)
        if not target:
            client.stdout(json_data_pretty)
        else:
            if os.path.isdir(target):
                raise FileExistsError('Please specify a file path. Target '
                                      'is a directory: {0}'.format(target))
            if not force and os.path.isfile(target):
                raise FileExistsError('Please specify a file that does '
                                      'not exist or use the force flag: '
                                      '{0} bdc config show -t '
                                      '{1} --force'.format(Configuration().CLI_NAME, target))
            folder_name = os.path.dirname(target)
            if folder_name:
                os.makedirs(os.path.dirname(target), exist_ok=True)

            with open(target, "w") as output:
                output.write(json_data_pretty)
    except Exception as e:
        raise CliError(e)


def bdc_debug_copy_logs(client,
                            namespace,
                            container=None,
                            target_folder=None,
                            pod=None,
                            timeout=0,
                            skip_compress=False,
                            exclude_dumps=False):
    """
    Copy Logs commands - requires kube config
    """
    try:
        # -- Check Kubectl Context --
        util.check_and_set_kubectl_context()

        copy_debug_logs(
            namespace, target_folder, pod, container, timeout, skip_compress, exclude_dumps)
    except ClusterError as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def bdc_debug_dump(
        client, namespace, container='controller', target_folder='./output/dump'):
    """
    Trigger dump for given container and copy out the dump file to given
    output folder
    """
    try:
        # -- Check Kubectl Context --
        util.check_and_set_kubectl_context()

        take_dump(namespace, container, target_folder)
    except ClusterError as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def bdc_config_list(client, config_profile=None, config_type=None, accept_eula=None):
    """
    Lists available configuration file choices.
    :param client:
    :param config_profile:
    :param config_type:
    :param accept_eula:
    :return:
    """
    try:
        # -- enforce EULA --
        client.profile.accept_eula = eula_prompt(client, accept_eula, True)

        if config_type and not config_profile:
            raise ValueError('Please specify a --config-profile when using --type.')

        if config_profile and not config_type:
            config_type = 'bdc'

        return DeploymentConfigUtil.config_list(CONFIG_DIR, config_profile, config_type)

    except ValueError as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def bdc_config_init(client, target=None, source=None, force=None, accept_eula=None):
    """
    Initializes a cluster configuration file for the user.
    :param client:
    :param target:
    :param source:
    :param force:
    :param accept_eula:
    :return:
    """
    try:
        # -- enforce EULA --
        client.profile.accept_eula = eula_prompt(client, accept_eula, True)

        try:
            if not target:
                target = prompt_for_input('Custom Config Profile Path:', 'custom', False, False)
        except NoTTYException:
            # If non-interactive, default to custom directory
            target = 'custom'

        stdout = client.stdout

        # Read the available configs by name
        config_map = DeploymentConfigUtil.get_config_map(CONFIG_DIR)

        if source:
            if source not in config_map.keys():
                raise ValueError('Invalid config source, please consult [bdc '
                                 'config list] for available sources')
        elif not source:
            config_names = DeploymentConfigUtil.get_config_display_names(config_map)
            # Prompt the user for a choice between configs
            stdout('Please choose a config:')
            source = prompt_for_choice(config_names, config_names[1])

        if os.path.isfile(target):
            raise FileExistsError('Please specify a directory path. Target is a file: {0}'.format(target))

        result = DeploymentConfigUtil.save_config_profile(target, source, CONFIG_DIR, CONFIG_FILES, config_map, force)

        return 'Created configuration profile in {}'.format(result)

    except ValueError as e:
        raise CliError(e)
    except NoTTYException:
        raise CliError('Please specify target and source in non-interactive mode')
    except Exception as e:
        raise CliError(e)


def bdc_config_add(client, config_file, json_values):
    """
    Add new key and value to the given config file
    :param client:
    :param config_file:
    :param json_values:
    :return:
    """
    try:
        config_object = DeploymentConfigUtil.config_add(config_file, json_values)
        DeploymentConfigUtil.write_config_file(config_file, config_object)
    except Exception as e:
        raise CliError(e)


def bdc_config_replace(client, config_file, json_values):
    """
    Replace the value of a given key in the given config file
    :param client:
    :param config_file:
    :param json_values:
    :return:
    """
    try:
        config_object = DeploymentConfigUtil.config_replace(config_file, json_values)
        DeploymentConfigUtil.write_config_file(config_file, config_object)
    except Exception as e:
        raise CliError(e)


def bdc_config_remove(client, config_file, json_path):
    """
    Remove a key from the given config file
    :param client:
    :param config_file:
    :param json_path:
    :return:
    """
    try:
        config_object = DeploymentConfigUtil.config_remove(config_file, json_path)
        DeploymentConfigUtil.write_config_file(config_file, config_object)
    except Exception as e:
        raise CliError(e)


def bdc_config_patch(client, config_file, patch_file):
    """
    Patch a given file against the given config file
    :param client:
    :param config_file:
    :param patch_file:
    :return:
    """
    try:
        config_object = DeploymentConfigUtil.config_patch(config_file, patch_file)
        DeploymentConfigUtil.write_config_file(config_file, config_object)
    except Exception as e:
        raise CliError(e)
