# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

import json
import os
import base64
import yaml
import sys
import time
import azdata.cli.core.deploy as util

from azdata.cli.core.logging import get_logger
from azdata.cli.core.constants import (
    AZDATA_USERNAME,
    AZDATA_PASSWORD,
    ARC_GROUP,
    ARC_API_VERSION,
    DATA_CONTROLLER_PLURAL)
from azdata.cli.core.prompt import (prompt, prompt_pass)
from azdata.cli.commands.sqlmi.exceptions import SqlmiError
from azdata.cli.commands.sqlmi.util import is_valid_sql_password
from azdata.cli.core.util import (
    FileUtil,
    is_windows)
from azdata.cli.core.clients.kubernetes_client import (
    K8sApiException)
from azdata.cli.core.models.data_controller_custom_resource import DataControllerCustomResource
from azdata.cli.core.models.custom_resource_definition import CustomResourceDefinition
from azdata.cli.commands.sqlmi.constants import (RESOURCE_KIND, RESOURCE_KIND_PLURAL, API_GROUP, API_VERSION)
from azdata.cli.commands.sqlmi.models.sqlmi_cr_model import (SqlmiCustomResource)
from azdata.cli.core.exceptions import (KubernetesError, CliError, http_status_codes)
from azdata.cli.core.deploy import DeploymentConfigUtil
from humanfriendly.terminal.spinners import AutomaticSpinner
from urllib3.exceptions import NewConnectionError, MaxRetryError

CONNECTION_RETRY_ATTEMPTS = 12
RETRY_INTERVAL = 5

logger = get_logger(__name__)


def arc_sql_mi_create(
        client,
        name,
        path=None,
        # namespace=None,
        cores_limit=None,
        cores_request=None,
        memory_limit=None,
        memory_request=None,
        storage_class_data=None,
        storage_class_logs=None,
        storage_class_datalogs=None,
        storage_class_backups=None,
        volume_size_data=None,
        volume_size_logs=None,
        volume_size_datalogs=None,
        volume_size_backups=None,
        no_external_endpoint=None,
        dev=None,
        # port=None,
        no_wait=False):
    """
    Create a SQL managed instance.
    :param client:
    :param path: The path to the src file for the SQL managed instance json file.
    :param name: The name of the Azure SQL managed instance.
    :param namespace: Namespace where the Azure SQL managed instance is to be deployed.
    :param cores_limit: The cores limit of the managed instance as an integer.
    :param cores_request: The request for cores of the managed instance as an integer.
    :param memory_limit: The limit of the capacity of the managed instance as an integer.
    :param memory_request: The request for the capacity of the managed instance as an integer amount of memory in GBs.
    :param storage_class_data: The storage classes to be used for data
    :param storage_class_logs: The storage classes to be used for logs (/var/log).
    :param storage_class_datalogs: The storage class to be used for data-logs (/var/opt/mssql/data-log)
    :param storage_class_backups: The storage classes to be used for logs (/var/opt/mssql/backups).
    :param volume_size_data: The volume size for the storage classes to be used for data.
    :param volume_size_logs: The volume size for the storage classes to be used for logs.
    :param volume_size_datalogs: The volume size for the storage classes to be used for data logs.
    :param volume_size_backups: The volume size for the storage classes to be used for backups.
    :param no_external_endpoint: If not specified, an external service is created using the same service type as dc.
    :param dev: If this is specified, then it is considered a dev instance and will not be billed for.
    :param port: Optional. Default is 31433.
    :param no_wait: If given, the command won't wait for the deployment to be ready before returning.
    :return:
    """
    try:
        util.check_and_set_kubectl_context()

        # Retrieve CRD
        #
        crd_object = util.retry(lambda: client.apis.controller.resource_crd_get(RESOURCE_KIND),
                                retry_count=CONNECTION_RETRY_ATTEMPTS,
                                retry_delay=RETRY_INTERVAL,
                                retry_method="resource crd get",
                                retry_on_exceptions=(NewConnectionError, MaxRetryError))

        crd = CustomResourceDefinition(crd_object)

        # TODO: Support user supplied namespace when the backend supports it
        namespace = client.profile.active_context.namespace

        # If no template files are provided,
        # retrieve them from the Arc Data Controller.
        if not path:
            spec_object = util.retry(lambda: client.apis.controller.resource_spec_get(RESOURCE_KIND),
                                     retry_count=CONNECTION_RETRY_ATTEMPTS,
                                     retry_delay=RETRY_INTERVAL,
                                     retry_method="resource spec get",
                                     retry_on_exceptions=(NewConnectionError, MaxRetryError))

            cr = SqlmiCustomResource(spec_object)
        # Otherwise, use the provided src file.
        else:
            json_object = FileUtil.read_json(path)
            cr = SqlmiCustomResource(json_object)

        # Name required
        cr.name = name
        cr.namespace = namespace

        custom_object_exists = util.retry(lambda: client.apis.kubernetes.namespaced_custom_object_exists(
                                                    cr.name, cr.namespace,
                                                    group=API_GROUP,
                                                    version=API_VERSION,
                                                    plural=RESOURCE_KIND_PLURAL),
                                          retry_count=CONNECTION_RETRY_ATTEMPTS,
                                          retry_delay=RETRY_INTERVAL, retry_method="get namespaced custom object",
                                          retry_on_exceptions=(NewConnectionError, MaxRetryError, KubernetesError))

        if custom_object_exists:
            raise ValueError("Arc SQL managed instance `{}` already exists in namespace `{}`.".format(name, namespace))

        if cores_limit:
            cr.cores_limit = cores_limit

        if cores_request:
            cr.cores_request = cores_request

        if memory_limit:
            cr.memory_limit = memory_limit

        if memory_request:
            cr.memory_request = memory_request

        # Validate cores/memory after both have been set to avoid behavior related to the order of property setting
        #
        cr.validate_cores()
        cr.validate_memory()

        if storage_class_data:
            if not client.apis.kubernetes.storage_class_exists(storage_class_data):
                raise ValueError("Storage class \"%s\" does not exist" % storage_class_data)
            cr.storage_class_data = storage_class_data

        if storage_class_logs:
            if not client.apis.kubernetes.storage_class_exists(storage_class_logs):
                raise ValueError("Storage class \"%s\" does not exist" % storage_class_logs)
            cr.storage_class_logs = storage_class_logs

        if storage_class_datalogs:
            cr.storage_class_datalogs = storage_class_datalogs

        if storage_class_backups:
            cr.storage_class_backups = storage_class_backups

        if volume_size_data:
            cr.volume_size_data = volume_size_data

        if volume_size_logs:
            cr.volume_size_logs = volume_size_logs

        if volume_size_datalogs:
            cr.volume_size_datalogs = volume_size_datalogs

        if volume_size_backups:
            cr.volume_size_backups = volume_size_backups

        # if collation:
        #     raise NotImplementedError('The collation arg is currently unimplemented.')

        if not no_external_endpoint:
            response = util.retry(lambda: client.apis.kubernetes.list_namespaced_custom_object(
                                            namespace,
                                            group=ARC_GROUP,
                                            version=ARC_API_VERSION,
                                            plural=DATA_CONTROLLER_PLURAL),
                                  retry_count=CONNECTION_RETRY_ATTEMPTS,
                                  retry_delay=RETRY_INTERVAL,
                                  retry_method="list namespaced custom object",
                                  retry_on_exceptions=(NewConnectionError, MaxRetryError, K8sApiException))

            dcs = response.get("items")
            if not dcs:
                raise CliError("No data controller exists in namespace `{}`. Cannot set external endpoint argument."
                               .format(namespace))
            else:
                dc_cr = DataControllerCustomResource(dcs[0])
                cr.service_type = dc_cr.controller_service_type

        # if port:
        #    cr.service_port = port

        if dev:
            cr.dev = dev

        # Create Custom Resource Definition
        #
        util.retry(lambda: client.apis.kubernetes.create_custom_resource_definition(crd),
                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                   retry_delay=RETRY_INTERVAL, retry_method="create custom resource definition",
                   retry_on_exceptions=(NewConnectionError, MaxRetryError, K8sApiException, KubernetesError))

        # Create login secret
        #
        secret_name = name + "-login-secret"

        secret_exists = util.retry(lambda: client.apis.kubernetes.secret_exists(cr.namespace, secret_name),
                                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                                   retry_delay=RETRY_INTERVAL, retry_method="secret exists",
                                   retry_on_exceptions=(NewConnectionError, MaxRetryError, K8sApiException))

        if not secret_exists:

            # Username
            username = os.environ.get(AZDATA_USERNAME)
            if not username:
                if sys.stdin.isatty():
                    username = prompt("Arc SQL managed instance username:")
                else:
                    raise ValueError("Please provide an Arc SQL managed instance password through the env "
                                     "variable AZDATA_USERNAME.")
            else:
                client.stdout("Using AZDATA_USERNAME environment variable for `{}` username.".format(name))

            while username == "sa":
                username = prompt("The login 'sa' is not allowed.  Please use a different login.")

            while username == "":
                username = prompt("Login username required. Please enter a login.")

            # Password
            pw = os.environ.get(AZDATA_PASSWORD)
            if not pw:
                if sys.stdin.isatty():
                    while not pw:
                        pw = prompt_pass("Arc SQL managed instance password:", True)
                        if not is_valid_sql_password(pw, "sa"):
                            client.stdout("\nError: SQL Server passwords must be at least 8 characters long, cannot "
                                            "contain the username, and must contain characters from three of the"
                                            " following four sets: Uppercase letters, Lowercase letters, Base 10 digits,"
                                            "and Symbols. Please try again.\n")
                            pw = None
                else:
                    raise ValueError("Please provide an Arc SQL managed instance password through the env "
                                     "variable AZDATA_PASSWORD.")
            else:
                client.stdout("Using AZDATA_PASSWORD environment variable for `{}` password.".format(name))

            secrets = dict()
            encoding = "utf-8"
            secrets["secretName"] = secret_name
            secrets["base64Username"] = base64.b64encode(
                bytes(username, encoding)).decode(encoding)
            secrets["base64Password"] = base64.b64encode(
                bytes(pw, encoding)).decode(encoding)
            temp = util.get_config_from_template(
                os.path.join(os.path.dirname(os.path.realpath(__file__)), "templates", "useradmin-login.yaml.tmpl"),
                secrets)
            mssql_secret = yaml.full_load(temp)

            try:
                util.retry(
                    lambda: client.apis.kubernetes.create_secret(cr.namespace, mssql_secret, ignore_conflict=True),
                    retry_count=CONNECTION_RETRY_ATTEMPTS,
                    retry_delay=RETRY_INTERVAL, retry_method="create secret",
                    retry_on_exceptions=(NewConnectionError, MaxRetryError, K8sApiException))

            except K8sApiException as e:
                if e.status != http_status_codes.conflict:
                    raise

        # Create custom resource
        #
        util.retry(lambda: client.apis.kubernetes.create_namespaced_custom_object(crd, cr, ignore_conflict=True),
                   retry_count=CONNECTION_RETRY_ATTEMPTS,
                   retry_delay=RETRY_INTERVAL, retry_method="create namespaced custom object",
                   retry_on_exceptions=(NewConnectionError, MaxRetryError, KubernetesError))

        if no_wait:
            client.stdout(
                "Deployed {0} in namespace `{1}`. Please use `azdata arc sql mi show -n {0}` to check its status."
                    .format(cr.name, cr.namespace))
        else:
            state = None
            if not is_windows():
                with AutomaticSpinner('Deploying {0} in namespace `{1}`'.format(cr.name, cr.namespace),
                                      show_time=True):
                    while state != 'ready' or state is None:
                        time.sleep(5)
                        response = util.retry(lambda: client.apis.kubernetes.get_namespaced_custom_object(
                                                        cr.name, cr.namespace,
                                                        group=API_GROUP,
                                                        version=API_VERSION,
                                                        plural=RESOURCE_KIND_PLURAL),
                                              retry_count=CONNECTION_RETRY_ATTEMPTS,
                                              retry_delay=RETRY_INTERVAL, retry_method="get namespaced custom object",
                                              retry_on_exceptions=(NewConnectionError, MaxRetryError, KubernetesError))

                        deployed_cr = SqlmiCustomResource(response)
                        state = deployed_cr.state
                        if state is not None:
                            state = state.lower()
            else:
                client.stdout('Deploying {0} in namespace `{1}`'.format(name, namespace))
                while state != 'ready' or state is None:
                    time.sleep(5)
                    response = util.retry(lambda: client.apis.kubernetes.get_namespaced_custom_object(
                                                    cr.name, cr.namespace,
                                                    group=API_GROUP,
                                                    version=API_VERSION,
                                                    plural=RESOURCE_KIND_PLURAL),
                                          retry_count=CONNECTION_RETRY_ATTEMPTS,
                                          retry_delay=RETRY_INTERVAL, retry_method="get namespaced custom object",
                                          retry_on_exceptions=(NewConnectionError, MaxRetryError, KubernetesError))

                    deployed_cr = SqlmiCustomResource(response)
                    state = deployed_cr.state
                    if state is not None:
                        state = state.lower()

            client.stdout('{0} is Ready'.format(cr.name))

    except KubernetesError as e:
        raise SqlmiError(e.message)
    except ValueError as e:
        raise CliError(e)
    except Exception as e:
        raise CliError(e)


def arc_sql_mi_edit(
        client,
        name,
        # namespace=None,
        path=None,
        cores_limit=None,
        cores_request=None,
        memory_limit=None,
        memory_request=None,
        dev=None,
        # port=None,
        no_wait=False):
    """
    Edit the configuration of a SQL managed instance.
    :param client:
    :param path: The path to the src file for the SQL managed instance json file.
    :param name: The name of the Azure SQL managed instance.
    :param namespace: Namespace where the Azure SQL managed instance is to be deployed.
    :param cores_limit: The cores limit of the managed instance as an integer.
    :param cores_request: The request for cores of the managed instance as an integer.
    :param memory_limit: The limit of the capacity of the managed instance as an integer.
    :param memory_request: The request for the capcity of the managed instance as an integer amount of memory in GBs.
    :param dev: If this is specified, then it is considered a dev instance and will not be billed for.
    :param port: Optional. Default is 31433.
    :param no_wait: If given, the command won't wait for the deployment to be ready before returning.
    :return:
    """
    try:
        util.check_and_set_kubectl_context()

        # Retrieve CRD
        crd = CustomResourceDefinition(
            client.apis.controller.resource_crd_get(RESOURCE_KIND)
        )

        # TODO: Support user supplied namespace when the backend supports it
        namespace = client.profile.active_context.namespace

        if path:
            # Read src file for edit
            json_object = FileUtil.read_json(path)
        else:
            json_object = client.apis.kubernetes.get_namespaced_custom_object(name, namespace, crd=crd)

        cr = SqlmiCustomResource(json_object)

        # Name required
        cr.name = name

        cr.namespace = namespace

        if cores_limit:
            cr.cores_limit = cores_limit

        if cores_request:
            cr.cores_request = cores_request

        if memory_limit:
            cr.memory_limit = memory_limit

        if memory_request:
            cr.memory_request = memory_request

        # Validate cores/memory after both have been set to avoid behavior related to the order of property setting
        #
        cr.validate_memory()
        cr.validate_cores()

        # if collation:
        #     raise NotImplementedError("The collation arg is currently unimplemented.")

        # if port:
        #     cr.service_port = port

        if dev:
            cr.dev = dev

        # Patch CR
        client.apis.kubernetes.patch_namespaced_custom_object(
            crd, cr
        )

        if no_wait:
            client.stdout(
                "Updated {0} in namespace `{1}`. Please use `azdata arc sql mi show -n {0} to check its status."
                    .format(cr.name, cr.namespace))
        else:
            state = None
            if not is_windows():
                with AutomaticSpinner('Updating {0} in namespace `{1}`'.format(cr.name, cr.namespace),
                                      show_time=True):
                    while state != 'ready' or state is None:
                        time.sleep(5)
                        response = client.apis.kubernetes.get_namespaced_custom_object(
                            cr.name, cr.namespace, group=API_GROUP, version=API_VERSION, plural=RESOURCE_KIND_PLURAL
                        )
                        deployed_cr = SqlmiCustomResource(response)
                        state = deployed_cr.state
                        if state is not None:
                            state = state.lower()
            else:
                client.stdout('Updating {0} in namespace `{1}`'.format(name, namespace))
                while state != 'ready' or state is None:
                    time.sleep(5)
                    response = client.apis.kubernetes.get_namespaced_custom_object(
                        cr.name, cr.namespace, group=API_GROUP, version=API_VERSION, plural=RESOURCE_KIND_PLURAL
                    )
                    deployed_cr = SqlmiCustomResource(response)
                    state = deployed_cr.state
                    if state is not None:
                        state = state.lower()

            client.stdout('{0} is Ready'.format(cr.name))

    except KubernetesError as e:
        raise SqlmiError(e.message)
    except Exception as e:
        raise CliError(e)


def arc_sql_mi_delete(client, name):
    """
    Delete a SQL managed instance.
    :param client:
    :param name: Name of the SQL managed instance.
    :return:
    """
    try:
        util.check_and_set_kubectl_context()

        # TODO: Support user supplied namespace when the backend supports it
        namespace = client.profile.active_context.namespace

        crd = CustomResourceDefinition(
            client.apis.controller.resource_crd_get(RESOURCE_KIND)
        )
        client.apis.kubernetes.delete_namespaced_custom_object(
            crd, name, namespace
        )

        client.stdout("Deleted {} from namespace {}".format(name, namespace))

    except KubernetesError as e:
        raise SqlmiError(e.message)
    except Exception as e:
        raise CliError(e)


def arc_sql_mi_show(client, name, path=None):
    """
    Show the details of a SQL managed instance.
    :param client:
    :param name: Name of the SQL managed instance.
    :param path: A path where the full specification for the SQL managed instance should be written.
    :return:
    """
    try:
        util.check_and_set_kubectl_context()

        # TODO: Support user supplied namespace when the backend supports it
        namespace = client.profile.active_context.namespace

        response = client.apis.kubernetes.get_namespaced_custom_object(
            name, namespace, group=API_GROUP, version=API_VERSION, plural=RESOURCE_KIND_PLURAL
        )

        if path:
            if not os.path.isdir(path):
                os.makedirs(path)
            path = os.path.join(path, "{}.json".format(name))
            with open(path, "w") as outfile:
                json.dump(response, outfile, indent=4)
            client.stdout("{0} specification written to {1}".format(name, path))
        else:
            return response

    except KubernetesError as e:
        raise SqlmiError(e.message)
    except Exception as e:
        raise CliError(e)


def arc_sql_mi_list(client):
    """
    List SQL managed instances.
    :param client:
    :return:
    """
    try:
        util.check_and_set_kubectl_context()

        # TODO: Support user supplied namespace when the backend supports it
        namespace = client.profile.active_context.namespace

        response = client.apis.kubernetes.list_namespaced_custom_object(
            namespace, group=API_GROUP, version=API_VERSION, plural=RESOURCE_KIND_PLURAL
        )

        items = response.get("items")

        result = []

        # Temporary, need to discuss what the intended structure is across partners
        #
        for item in items:
            cr = SqlmiCustomResource(item)
            result.append(
                {"name": cr.name,
                 "externalEndpoint": cr.external_endpoint,
                 "replicas": cr.ready_replicas,
                 "state": cr.state})

        return result

    except KubernetesError as e:
        raise SqlmiError(e.message)
    except Exception as e:
        raise CliError(e)


def arc_sql_mi_config_init(client, path):
    """
    Returns a package of crd.json and spec-template.json.
    :param client:
    :param path:
    :return:
    """
    try:
        if not os.path.isdir(path):
            os.makedirs(path, exist_ok=True)

        if not is_windows():
            with AutomaticSpinner("Fetching {0} template".format(RESOURCE_KIND),
                                  show_time=True):
                crd_response = client.apis.controller.resource_crd_get(RESOURCE_KIND)
                spec_response = client.apis.controller.resource_spec_get(RESOURCE_KIND)
        else:
            crd_response = client.apis.controller.resource_crd_get(RESOURCE_KIND)
            spec_response = client.apis.controller.resource_spec_get(RESOURCE_KIND)

        crd_pretty = json.dumps(crd_response, indent=4)
        spec_pretty = json.dumps(spec_response, indent=4)

        with open(os.path.join(path, "crd.json"), "w") as output:
            output.write(crd_pretty)

        with open(os.path.join(path, "spec.json"), "w") as output:
            output.write(spec_pretty)

        client.stdout("{0} template created in directory: {1}".format(RESOURCE_KIND, path))

    except Exception as e:
        raise CliError(e)


def arc_sql_mi_config_add(client, path, json_values):
    """
    Add new key and value to the given config file
    :param client:
    :param path:
    :param json_values:
    :return:
    """
    try:
        config_object = DeploymentConfigUtil.config_add(path, json_values)
        DeploymentConfigUtil.write_config_file(path, config_object)
    except Exception as e:
        raise CliError(e)


def arc_sql_mi_config_replace(client, path, json_values):
    """
    Replace the value of a given key in the given config file
    :param client:
    :param path:
    :param json_values:
    :return:
    """
    try:
        config_object = DeploymentConfigUtil.config_replace(path, json_values)
        DeploymentConfigUtil.write_config_file(path, config_object)
    except Exception as e:
        raise CliError(e)


def arc_sql_mi_config_remove(client, path, json_path):
    """
    Remove a key from the given config file
    :param client:
    :param path:
    :param json_path:
    :return:
    """
    try:
        config_object = DeploymentConfigUtil.config_remove(path, json_path)
        DeploymentConfigUtil.write_config_file(path, config_object)
    except Exception as e:
        raise CliError(e)


def arc_sql_mi_config_patch(client, path, patch_file):
    """
    Patch a given file against the given config file
    :param client:
    :param path:
    :param patch_file:
    :return:
    """
    try:
        config_object = DeploymentConfigUtil.config_patch(path, patch_file)
        DeploymentConfigUtil.write_config_file(path, config_object)
    except Exception as e:
        raise CliError(e)
