# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.core.logging import get_logger
from azdata.cli.core.exceptions import KerberosExchangeError

try:
    import kerberos
except ImportError:
    import winkerberos as kerberos

logger = get_logger(__name__)


class KerberosClient(object):
    """
    The client for managing
    """
    def __init__(self, endpoint, principal=None):
        self.endpoint = endpoint
        self.principal = principal
        self.context = None

    @property
    def endpoint(self):
        """
        Gets the endpoint used during L{authGSSClientInit}.
        """
        return self._endpoint

    @endpoint.setter
    def endpoint(self, endpoint):
        """
        Sets the endpoint used during L{authGSSClientInit}.
        :param endpoint: The endpoint.
        """
        self._endpoint = endpoint

    @property
    def context(self):
        """
        Gets the context object returned from L{authGSSClientInit}.
        """
        return self._context

    @context.setter
    def context(self, context):
        """
        Sets the context object.
        :param context: The context object returned from L{authGSSClientInit}.
        """
        self._context = context

    @property
    def principal(self):
        return self._principal

    @principal.setter
    def principal(self, principal):
        self._principal = principal

    def acquire_gssapi_token(self):
        """
        GSSAPI client-side authentication:

        - Initializes a context for GSSAPI client-side authentication.
        - Processes a single GSSAPI client-side step using the supplied data.
        - Process client response and attempt to retrieve ticket.

        :return: A `str` containing the client data to be sent to the server.
        """

        # -- 1. Initializes a context for GSSAPI client-side authentication --
        stage = 'authGSSClientInit'
        try:
            logger.debug('KERBEROS: stage %s', stage)

            flags = kerberos.GSS_C_MUTUAL_FLAG | kerberos.GSS_C_SEQUENCE_FLAG

            urlparse = self.endpoint.split('://')
            host = urlparse[0] if len(urlparse) == 1 else urlparse[1]

            service = 'HTTP@{0}'.format(host)
            logger.debug('KERBEROS: service %s', service)

            if self.principal:
                logger.debug('KERBEROS: principle %s',
                             self.principal.split(':')[0])

            code, self.context = kerberos.authGSSClientInit(
                service,
                gssflags=flags,
                principal=self.principal
            )

            logger.debug('KERBEROS: code %s', code)
            logger.debug('KERBEROS: context -----------')
            logger.debug(self.context)
            logger.debug('-----------------------------')

            if code < kerberos.AUTH_GSS_COMPLETE:
                msg = 'KERBEROS: Could not initialize at stage "{}"'.format(
                    stage)
                logger.debug(msg)
                raise KerberosExchangeError(msg, status=code)
        except (kerberos.GSSError, Exception) as e:
            msg = 'Could not initialize at stage "{}"\n{}'.format(stage, e)
            logger.debug(msg)
            raise KerberosExchangeError(msg)

        # -- 2. Processes a single GSSAPI client-side step --
        stage = 'authGSSClientStep'
        try:
            logger.debug('KERBEROS: stage %s', stage)
            code = kerberos.authGSSClientStep(self.context, '')

            logger.debug('KERBEROS: code %s', code)

            if code < kerberos.AUTH_GSS_CONTINUE:
                msg = 'KERBEROS: Client step failed at stage "{}"'.format(stage)
                logger.debug(msg)
                raise KerberosExchangeError(msg, status=code)
        except (kerberos.GSSError, Exception) as e:
            msg = 'Client step failed at stage "{}"\n{}'.format(stage, e)
            logger.debug(msg)
            raise KerberosExchangeError(msg)

        # -- 3. Process response --
        try:
            stage = 'authGSSClientResponse'
            logger.debug('KERBEROS: stage %s', stage)
            return kerberos.authGSSClientResponse(self.context)
        except (kerberos.GSSError, Exception) as e:
            msg = 'Client response failed at stage "{}"\n{}'.format(stage, e)
            logger.debug(msg)
            raise KerberosExchangeError(msg)


