# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

import os
import re
from enum import Enum

BASE = os.path.dirname(os.path.realpath(__file__))
"""
Base directory
"""

IS_WINDOWS = os.name == 'nt'
"""
Boolean for whether the CLI is in the Windows env
"""

IO_DELIM = '(?<!\\\\),(?=[\w/.$]+)'
"""
Defines the IO delimiter for json key/value pairs on config set
"""

KEY_VALUE_SPLIT = '(?<!\\\\)(?<!=)=(?!=)'
"""
Defines the key/value delimiter
"""

PLATFORM_WINDOWS = 'windows'
"""
Defines the Windows platform system name.
"""

AZDATA_USERNAME = 'AZDATA_USERNAME'
"""
Defines the username env variable for login
"""

AZDATA_PASSWORD = 'AZDATA_PASSWORD'
"""
Defines the password env variable for login
"""

DOCKER_USERNAME = 'DOCKER_USERNAME'
"""
Defines the username env variable for docker private registries.
"""

DOCKER_PASSWORD = 'DOCKER_PASSWORD'
"""
Defines the password env variable for docker private registries.
"""

REGISTRY_USERNAME = 'REGISTRY_USERNAME'
"""
Alternative definition for the username env variable for docker private registries.
"""

REGISTRY_PASSWORD = 'REGISTRY_PASSWORD'
"""
Alternative definition for the password env variable for docker private registries.
"""

WHEEL_INFO_RE = re.compile(
    r"""^(?P<namever>(?P<name>.+?)(-(?P<ver>\d.+?))?)
    ((-(?P<build>\d.*?))?-(?P<pyver>.+?)-(?P<abi>.+?)-(?P<plat>.+?)
    \.whl|\.dist-info)$""",
    re.VERBOSE).match
"""
Tokenize a Python wheel's filename.

{distro-{version}(-{build-tag})?-{python-tag}-{abi tag}-{platform-tag}.whl

Example:

requests-9.0.1-py2.py3-none-any.whl 

The version 9.0.1 build of the requests package is compatible with both Python 2 
and 3, and has no Application Binary Interface (pure Python), and is compatible 
with any CPU architecture.
"""

SEMANTIC_VERSION_RE = re.compile(
    r"""
        ^
        (?P<major>0|[1-9]\d*)
        \.
        (?P<minor>0|[1-9]\d*)
        \.
        (?P<patch>0|[1-9]\d*)
        (?:-(?P<prerelease>
        (?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)
        (?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))
        ?
        (?:\+(?P<buildmetadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$
    """,
    re.VERBOSE
)
"""
Regular expression of identifying <MAJOR.MINOR.PATCH> out of a string.

`https://semver.org/` provided the regx for guidance

Additional labels for pre-release and build metadata are available as 
extensions to the MAJOR.MINOR.PATCH format.

A pre-release version MAY be denoted by appending a hyphen and a series of dot 
separated identifiers immediately following the patch version. Identifiers MUST 
comprise only ASCII alphanumerics and hyphens [0-9A-Za-z-]. Identifiers MUST NOT 
be empty. Numeric identifiers MUST NOT include leading zeroes. Pre-release 
versions have a lower precedence than the associated normal version. A 
pre-release version indicates that the version is unstable and might not 
satisfy the intended compatibility requirements as denoted by its associated 
normal version. 

Examples: 
1.0.0-alpha, 1.0.0-alpha.1, 1.0.0-0.3.7, 1.0.0-x.7.z.92, 1.0.0-x-y-z.–.

Build metadata MAY be denoted by appending a plus sign and a series of dot 
separated identifiers immediately following the patch or pre-release version. 
Identifiers MUST comprise only ASCII alphanumerics and hyphens [0-9A-Za-z-]. 
Identifiers MUST NOT be empty. Build metadata MUST be ignored when determining 
version precedence. Thus two versions that differ only in the build metadata, 
have the same precedence. 

Examples: 
1.0.0-alpha+001, 1.0.0+20130313144700, 1.0.0-beta+exp.sha.5114f85, 
1.0.0+21AF26D3—-117B344092BD.
"""

# ------------------------------------------------------------------------------
# Arc constants for convenience of not rereading data controller CRD file for
# multiple API calls across command modules
# ------------------------------------------------------------------------------
ARC_GROUP = 'arcdata.microsoft.com'
"""
Defines the group for Arc CRDs
"""

ARC_API_VERSION = 'v1alpha1'
"""
Defines the kubernetes api version for Arc CRDs
"""

KUBERNETES_LABEL_PREFIX = 'app.kubernetes.io/'
"""
Defines the prefix for common Kubernetes labels
https://kubernetes.io/docs/concepts/overview/working-with-objects/common-labels/
"""

ARC_INSTANCE_LABEL = KUBERNETES_LABEL_PREFIX + 'instance'
"""
Defines the instance label for Arc resources
"""

ARC_RESOURCE_KIND_LABEL = KUBERNETES_LABEL_PREFIX + 'part-of'
"""
Defines the resource kind label for Arc resources
"""

DATA_CONTROLLER_PLURAL = 'datacontrollers'
"""
Defines the plural name of data controllers.
"""

MGMT_PROXY = 'mgmtproxy-svc-external'
"""
Name of management proxy service
"""


# -----------------------------------------------------------------------------------------------------------------
# Helpful constants and Enums for dealing with kubernetes resource quantities
# -----------------------------------------------------------------------------------------------------------------

# SI Suffixes
ONE_k = 10 ** 3
ONE_M = 10 ** 6
ONE_G = 10 ** 9
ONE_T = 10 ** 12
ONE_P = 10 ** 15
ONE_E = 10 ** 18

# Base two equivalents
ONE_Ki = 2 ** 10
ONE_Mi = 2 ** 20
ONE_Gi = 2 ** 30
ONE_Ti = 2 ** 40
ONE_Pi = 2 ** 50
ONE_Ei = 2 ** 60

# Millicores
MILLICORES_PER_CORE = 10 ** 3

# A table for converting SI suffixes or their base two equivalents
# into their sizes in bytes
UNIT_TABLE = {
    "k": ONE_k, "Ki": ONE_Ki,
    "M": ONE_M, "Mi": ONE_Mi,
    "G": ONE_G, "Gi": ONE_Gi,
    "T": ONE_T, "Ti": ONE_Ti,
    "P": ONE_P, "Pi": ONE_Pi,
    "E": ONE_E, "Ei": ONE_Ei
}

# Helper regexps for numbers
# NOTE: cannot format these as regex since python automatically adds anchors to them
#
INTEGER = r"[0-9]+"
STRICT_POS_INTEGER = r"[0-9]*[1-9]+[0-9]*"
FLOAT = r"[0-9]*\.[0-9]+"
STRICT_POS_FLOAT = r"({}\.[0-9]*)|([0-9]*\.{})".format(STRICT_POS_INTEGER, STRICT_POS_INTEGER)
NUMBER = r"{}|{}".format(INTEGER, FLOAT)
STRICT_POS_NUMBER = r"({})|({})".format(STRICT_POS_FLOAT, STRICT_POS_INTEGER)

# Maximum decimals for kubernetes resource quantity
KUBE_QUANTITY_PRECISION = 3


class ResourceType(Enum):
    """
    Represents the supported resource request/limit types for azdata
    """
    MEMORY = 'memory'
    CPU = 'cpu'

    def __eq__(self, other):
        return self.value == other.value


class MemoryFormat(Enum):
    """
    The possible memory formats for a kubernetes memory quantity
    """
    SI = r"^[+]?((?P<value>{})(?P<suffix>[kMGTPE]))$".format(NUMBER)
    BASE_TWO = r"^[+]?((?P<value>{})(?P<suffix>[KMGTPE]i))$".format(NUMBER)
    SCIENTIFIC = r"^[+]?((?P<value>{})([eE])(?P<sign>[+-]?)(?P<exponent>{}))$".format(NUMBER, INTEGER)
    INT = r"^{}$".format(STRICT_POS_INTEGER)
    ALL = r"^(([+]?(({})|({}))([MGTPE]i?|Ki|k|e{})))|({})$".format(INTEGER, FLOAT, INTEGER, STRICT_POS_INTEGER)

    def __eq__(self, other):
        return self.value == other.value


class CpuFormat(Enum):
    """
    The possible formats for a kubernetes CPU quantity
    """
    WHOLE = r"^(?P<value>{})$".format(STRICT_POS_NUMBER)
    # Note precision higher than 1m is not allowed so we use Integer here
    MILLICORES = r"^(?P<value>{})m$".format(STRICT_POS_INTEGER)
    ALL = r"^(({})|({})m)$".format(STRICT_POS_NUMBER, STRICT_POS_INTEGER)

    def __eq__(self, other):
        return self.value == other.value
