# ------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ------------------------------------------------------------------------------

from azdata.cli.core.configuration import Configuration
from azdata.cli.core.text import Text
from azdata.cli.core.util import (FileUtil, semversion_parse)
from azdata.cli.core.logging import get_logger
from six import add_metaclass
from abc import (ABCMeta, abstractmethod)
from importlib import import_module

import os

__all__ = ['WheelExtension', 'PackageExtension']

logger = get_logger(__name__)


@add_metaclass(ABCMeta)
class BaseExtension(object):
    """
    Designed to be a low-level abstract foundation class from which other
    extension attribute classes can be derived.
    """

    def __init__(self, name, path=None):
        self._name = name
        self._path = path
        self._reload = False

    @property
    def name(self):
        return self._name

    @property
    def path(self):
        return self._path

    @property
    def reload(self):
        """
        Get whether to reload the extension module import on the `sys.path`.
        """
        return self._reload

    @reload.setter
    def reload(self, reload):
        """
        Set whether to reload the extension module import on the `sys.path`.
        """
        self._reload = reload

    @abstractmethod
    def preview(self):
        """Sub-class should override and implement."""

    @abstractmethod
    def version(self):
        """Sub-class should override and implement."""

    @abstractmethod
    def metadata(self):
        """Sub-class should override and implement."""

    @abstractmethod
    def module_name(self):
        """Sub-class should override and implement."""

    def is_eligible(self):
        """
        Test if this extension can be used in this CLI runtime.

        :return: `True` if this extension is eligible, `False` otherwise.
        """
        config = Configuration()
        cli_version = semversion_parse(config.version)
        ctrl_version = semversion_parse(config.server_version)

        meta = self.metadata
        min_cli = semversion_parse(meta.get('minCliVersion', None))
        max_cli = semversion_parse(meta.get('maxCliVersion', None))
        min_ctrl = semversion_parse(meta.get('minControllerVersion', None))
        max_ctrl = semversion_parse(meta.get('maxControllerVersion', None))
        el = {
            'cli_min': True,
            'cli_max': True,
            'ctrl_min': True,
            'ctrl_max': True
        }

        logger.debug('CLI Version: %s', cli_version)
        logger.debug('Controller Version: %s', ctrl_version)
        logger.debug('Minimum CLI Version: %s', min_cli)
        logger.debug('Maximum CLI Version: %s', max_cli)
        logger.debug('Minimum Controller Version: %s', min_ctrl)
        logger.debug('Maximum Controller Version: %s', max_ctrl)

        # -- min cli --
        cli_major_minor = \
            float(str(cli_version['major']) + '.' + str(cli_version['minor']))

        if min_cli:
            min_major_minor = \
                float(str(min_cli['major']) + '.' + str(min_cli['minor']))

            c = cli_major_minor >= min_major_minor
            el['cli_min'] = c and cli_version['patch'] >= min_cli['patch']

        # -- max cli --
        if max_cli:
            max_major_minor = \
                float(str(max_cli['major']) + '.' + str(max_cli['minor']))

            c = cli_major_minor <= max_major_minor
            el['cli_max'] = c and cli_version['patch'] <= max_cli['patch']

        # -- min controller --
        ctrl_major_minor = \
            float(str(ctrl_version['major']) + '.' + str(ctrl_version['minor']))

        if min_ctrl:
            min_major_minor = \
                float(str(min_ctrl['major']) + '.' + str(min_ctrl['minor']))

            c = ctrl_major_minor >= min_major_minor
            el['ctrl_min'] = c and ctrl_version['patch'] >= min_ctrl['patch']

        # -- max controller --
        if max_ctrl:
            max_major_minor = \
                float(str(max_ctrl['major']) + '.' + str(max_ctrl['minor']))

            c = ctrl_major_minor <= max_major_minor
            el['ctrl_max'] = c and ctrl_version['patch'] <= max_ctrl['patch']

        msg = 'The "{ext}" extension is not compatible with this ' \
              'version of the CLI.\n{const}Please install a ' \
              'compatible extension version or remove it.\n'

        if not el['cli_min']:
            constraint = 'You have CLI version ' \
                         '{v} and this extension requires a min of {min}.\n'.\
                format(v=config.version, min=meta.get('minCliVersion'))

            Text.warning(msg.format(ext=self.name, const=constraint))
            return False

        if not el['cli_max']:
            constraint = 'You have CLI version ' \
                         '{v} and this extension requires a max of {max}.\n'.\
                format(v=config.version, max=meta.get('maxCliVersion'))

            Text.warning(msg.format(ext=self.name, const=constraint))
            return False

        if not el['ctrl_min']:
            constraint = 'You have controller version ' \
                         '{v} and this extension requires a min\ncontroller ' \
                         '{min}.\n'.format(v=config.server_version,
                                           min=meta.get('minControllerVersion'))

            Text.warning(msg.format(ext=self.name, const=constraint))
            return False

        if not el['ctrl_max']:
            constraint = 'You have controller version ' \
                         '{v} and this extension requires a max\ncontroller ' \
                         '{max}.\n'.format(v=config.server_version,
                                           max=meta.get('maxControllerVersion'))

            Text.warning(msg.format(ext=self.name, const=constraint))
            return False

        return True

    def to_dict(self):
        """
        Convert the Extension state to a dictionary.
        """

        return {
            'name': self.name,
            'module': self.module_name,
            'version': self.version,
            'path': self.path or '<pre-installed>',
            'preview': self.preview,
            'reload': self.reload
        }

    def __str__(self):
        """
        Extension ToString.
        """
        return '\n' + '\n'.join([
            'Extension name "{}"'.format(self.name),
            'Extension directory "{}"'.format(self.path or '<pre-installed>'),
            'Extension in preview "{}"'.format(str(self.preview)),
            'Extension module name "{}"'.format(self.module_name),
            'Extension reload "{}"'.format(self.reload)
        ])

    @staticmethod
    def is_installed(ext_name):
        ext = BaseExtension.get_extension(ext_name)
        return ext and ext.path

    @staticmethod
    def get_extension(name):
        """
        Returns wheel-based extension.
        """
        for ext in BaseExtension.get_extensions():
            if ext.name == name:
                return ext
        return {}

    @staticmethod
    def get_extensions():
        """
        Returns all wheel-based extensions.
        """
        from azdata.cli.core.configuration import Configuration
        from glob import glob
        import os

        def _get_pre_package_extensions():
            """Load (optional) shipped 'extensions by package'."""
            ext_mods = dict()
            pkgs = []

            try:
                # If present load it other wise skip (optional)
                # Expected: `PACKAGE_EXTENSIONS = 'ext, ext2, ext3'`
                mod = import_module('azdata.cli.core.__extensions__')
                pkgs = [x.strip() for x in mod.PACKAGE_EXTENSIONS.split(',')]

                logger.debug(mod)
                logger.debug(pkgs)
            except (ImportError, Exception) as ex:
                logger.debug(ex)
                logger.debug('Skipping pre-package extensions, not found.')
                pass

            for name in pkgs:
                try:
                    ext = PackageExtension(name)
                    ext_mods[ext.name] = ext
                except (ImportError, AttributeError, Exception) as ex:
                    logger.debug(ex)
                    logger.debug('Skipping pre-package extension "%s"', name)
                    pass

            return ext_mods

        config = Configuration()
        ext_dir = config.extension_house

        # -- 1. Load any pre-installed package extensions first --
        exts = _get_pre_package_extensions()

        # -- 2. Load extensions installed by wheels from ext wheel house next --

        # check if extension was installed in edit mode `pip install -e`
        env_value = os.getenv('{}_EXTENSION_DIR'.
                              format(config.CLI_ENV_VAR_PREFIX.upper()))
        default_value = env_value or os.path.join(
            config.CLI_CONFIG_DIR, 'extensions')
        edit_mode = ext_dir != env_value and ext_dir != default_value

        logger.debug('ext_dir %s ', ext_dir)
        logger.debug('env_value %s ', env_value)
        logger.debug('default_value %s ', default_value)
        logger.debug('edit_mode %s', edit_mode)

        # By design this takes precedence over the above "extensions by
        # package". The reason for this is any first class preview extension
        # optionally shipped as part of the installers can be overridden post
        # install with a new extension version. By doing this here in this order
        # we can enforce this by simply replacing ext by name `exts['ext_name']`
        # should they match.
        if os.path.isdir(ext_dir):
            for ext_name in os.listdir(ext_dir):
                ext_path = os.path.join(ext_dir, ext_name)
                pattern = os.path.join(ext_path, '*.*-info')
                if os.path.isdir(ext_path) and glob(pattern):
                    ext = WheelExtension(ext_name, ext_path, edit_mode)
                    if ext_name in exts:
                        ext.reload = True
                    exts[ext_name] = ext

        return exts.values()

    @staticmethod
    def get_extension_path(ext_name):
        """
        Returns the the extension file-path from name.
        :param ext_name: The extension name.
        :return: The extension file-path to its directory location.
        """
        from azdata.cli.core.configuration import Configuration
        return os.path.join(Configuration().extension_house, ext_name)

    @staticmethod
    def get_extension_name_from_wheel(source):
        """
        Parses a full-path `source` wheel filename to extract the section
        portion that represents the name of the module.
        :param source: The wheel file.
        :return: The extension name.
        """
        from azdata.cli.core.constants import WHEEL_INFO_RE

        whl_filename = os.path.basename(source)
        return WHEEL_INFO_RE(whl_filename).groupdict().get('name'). \
            replace('_', '-')


# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class WheelExtension(BaseExtension):
    """
    "A `BaseExtension` derivative class defining an extension by `wheel`.
    This also reads values from its .dist-info/METADATA.
    """
    def __init__(self, name, path, edit_mode=False):
        super(WheelExtension, self).__init__(name, path)

        self._version = None
        self._module_name = None
        self._metadata = None
        self._edit_mode = edit_mode

    @property
    def preview(self):
        return self.metadata.get('isPreview', False)

    @property
    def edit_mode(self):
        return self._edit_mode

    @property
    def version(self):
        from glob import glob
        from azdata.cli.core.util import FileUtil
        import os

        if self._version:
            return self._version

        dist_info = ''
        version_token = 'Version:'

        try:
            for dist_info_dir in glob(os.path.join(self.path, '*.*-info')):
                if self.name.replace('-', '_') in dist_info_dir:
                    # 1. look in `metadata.json`
                    dist_info = os.path.join(dist_info_dir, 'metadata.json')

                    # 2. fallback to `METADATA`
                    if not os.path.exists(dist_info):
                        dist_info = os.path.join(dist_info_dir, 'METADATA')

                    # 2. fallback to `PKG-INFO`
                    if not os.path.exists(dist_info):
                        dist_info = os.path.join(dist_info_dir, 'PKG-INFO')
                    break

            if os.path.exists(dist_info):
                logger.debug('dist-info metadata found: %s', dist_info)
                if dist_info.endswith('.json'):
                    meta = FileUtil.read_json(dist_info)
                    self._version = meta.get('version', '')
                else:
                    # read METADATA|PKG-INFO line-by-line looking for `Version:`
                    for line in FileUtil.read_text_file(dist_info).split('\n'):
                        if line.startswith(version_token):
                            self._version = line.split(version_token)[1].strip()
                            break
            else:
                logger.debug('No dist-info metadata found for extension "{0}", '
                             'skipping...'.format(self.name))

        except Exception:
            logger.warning('Failed to read version in "%s"', dist_info)

        return self._version

    @property
    def metadata(self):
        import json
        import traceback

        if self._metadata:
            return self._metadata

        self._metadata = {}

        ext_metadata_path = os.path.join(
            self.path, self.module_name, 'azdata-metadata.json')

        if os.path.exists(ext_metadata_path):
            try:
                self._metadata = FileUtil.read_json(ext_metadata_path)
            except json.decoder.JSONDecodeError as ex:
                logger.warn('Malformed metadata json: \n%s', ext_metadata_path)
                logger.debug('Unable to load extension metadata: %s',
                             traceback.format_exc())

        return self._metadata

    @property
    def module_name(self):
        from azdata.cli.core.configuration import Configuration

        if self._module_name:
            return self._module_name

        prefix = Configuration().EXTENSIONS_PREFIX
        pos_mods = [n for n in os.listdir(self.path) if n.startswith(prefix)
                    and os.path.isdir(os.path.join(self.path, n))]

        if len(pos_mods) != 1:
            raise AssertionError('Expected 1 module to load starting with '
                                 '"{}": got {}'.format(prefix, pos_mods))

        self._module_name = pos_mods[0]

        return self._module_name

# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #
# ---------------------------------------------------------------------------- #


class PackageExtension(BaseExtension):
    """
    "A `BaseExtension` derivative class defining an extension by `package`.
    """
    def __init__(self, module_name):
        mod_attrs = import_module('{}.__version__'.format(module_name))
        super(PackageExtension, self).__init__(mod_attrs.__title__)

        self._module_name = module_name
        self._version = mod_attrs.__version__
        self._metadata = {'isPreview': True}  # TODO

    @property
    def preview(self):
        return self.metadata.get('isPreview')

    @property
    def version(self):
        return self._version

    @property
    def metadata(self):
        return self._metadata

    @property
    def module_name(self):
        return self._module_name
