import re
from azdata.cli.core.models.kube_resource import (
    MemoryResource,
    CpuResource)

class CustomResource(object):
    """
    Internal Custom Resource object to be used for deployments.
    """
    def __init__(self, cr_object):
        """
        Initializes a CR object with the given json.
        :param cr_object: The custom resource json.
        """
        self.body = cr_object

    @property
    def body(self):
        """
        Gets the cr object used for deployments.
        """
        return self._body

    @body.setter
    def body(self, cr_body):
        """
        Sets the cr json body used for deployments.
        :param cr_body: The custom resource json body.
        """
        self._body = cr_body

    @property
    def spec(self):
        """
        Gets the spec of this cr.
        :return:
        """
        return self.body.get("spec")

    @spec.setter
    def spec(self, spec_body):
        """
        Sets the spec of this cr.
        :param spec_body:
        :return:
        """
        self.body["spec"] = spec_body

    @property
    def metadata(self):
        """
        Returns the metadata of the cr.
        :return:
        """
        return self.body.get("metadata")

    @property
    def api(self):
        """
        Returns the api information of the cr.
        :return:
        """
        return self.body.get("apiVersion")

    @property
    def group(self):
        """
        Returns the api group of the cr.
        :return:
        """
        return self.api.split("/")[0]

    @property
    def version(self):
        """
        Returns the api version of the cr.
        :return:
        """
        return self.api.split("/")[1]

    @property
    def name(self):
        """
        Returns the name of the cr.
        :return:
        """
        return self.metadata.get("name")

    @name.setter
    def name(self, n):
        """
        Sets the name of the cr.
        :param n: Name of the cr.
        :return:
        """
        self.metadata["name"] = n

    @property
    def namespace(self):
        """
        Returns the namespace of the cr.
        :return:
        """
        return self.metadata.get("namespace")

    @namespace.setter
    def namespace(self, ns):
        """
        Sets the namespace of the cr.
        :param ns: Namespace of the cr.
        :return:
        """
        self.metadata["namespace"] = ns

    @property
    def status(self):
        """
        Returns the status of the cr.
        :return:
        """
        return self.body.get("status")

    @property
    def state(self):
        """
        Returns the state of the cr.
        :return:
        """
        if self.status:
            return self.status.get("state")
        else:
            return None

    @property
    def ready_replicas(self):
        """
        Returns the ready replicas of the cr.
        :return:
        """
        if self.status:
            return self.status.get("readyReplicas")
        else:
            return None

    @property
    def external_endpoint(self):
        """
        Returns the status external endpoint.
        :return:
        """
        if self.status:
            return self.status.get("externalEndpoint")
        else:
            return None

    @property
    def service(self):
        """
        Returns the service property.
        :return:
        """
        return self.spec.get("service")

    @service.setter
    def service(self, s):
        """
        Sets the service property
        :param s:
        :return:
        """
        self.spec["service"] = s

    @property
    def service_type(self):
        """
        Returns the service type property
        :return:
        """
        return self.service.get("type")

    @service_type.setter
    def service_type(self, s):
        """
        Sets the service type property.
        :param s:
        :return:
        """
        if s != "NodePort" and s != "LoadBalancer":
            raise ValueError("Service type must be set to either NodePort or LoadBalancer.")

        if self.service:
            self.service["type"] = s
        else:
            self.spec["service"] = {"type": s}

    @property
    def service_port(self):
        """
        Returns the service port property
        :return:
        """
        return self.service.get("port")

    @service_port.setter
    def service_port(self, p):
        """
        Sets the service port property.
        :param p:
        :return:
        """
        try:
            val = int(p)
            if val <= 0 or val >= 65535:
                raise ValueError()
        except ValueError:
            raise ValueError("Service port must be a positive integer, with a minimum of 1.")
        if self.service:
            self.service["port"] = val
        else:
            self.spec["service"] = {"port": val}

    @property
    def dev(self):
        """
        Returns the dev property.
        :return:
        """
        return self.spec.get("dev")

    @dev.setter
    def dev(self, d):
        """
        Sets the dev property
        :param d: boolean
        :return:
        """
        self.spec["dev"] = d

    @property
    def storage(self):
        """
        Returns the storage property.
        :return:
        """
        return self.spec.get("storage")

    @property
    def storage_data(self):
        """
        Returns the data volume property.
        :param self:
        :return:
        """
        if self.storage:
            return self.storage.get("data")
        else:
            return None

    @property
    def storage_class_data(self):
        """
        Returns the storage class of the data volume
        :param self:
        :return:
        """
        if self.storage_data:
            return self.storage_data.get("className")
        else:
            return None

    @storage_class_data.setter
    def storage_class_data(self, sc):
        """
        Sets the storage class of the data volume.
        :param self:
        :param sc: storage class
        :return:
        """
        if self.storage_data:
            self.storage_data["className"] = sc
        elif self.storage:
            self.storage["data"] = {"className": sc}
        else:
            self.spec["storage"] = {"data": {"className": sc}}

    @property
    def volume_size_data(self):
        """
        Returns the volume size of the data volume
        :param self:
        :return:
        """
        if self.storage_data:
            return self.storage_data.get("size")
        else:
            return None

    @volume_size_data.setter
    def volume_size_data(self, vs):
        """
        Sets the volume size of the data volume.
        :param self:
        :param sc: storage class
        :return:
        """
        val = re.search(r"^([1-9]+[0-9]*)(Gi|Mi|Ki)$", vs)
        if not val:
            raise ValueError("Volume size must be a positive integer ending in `Ki`, `Mi`, or `Gi`.")

        if self.storage_data:
            self.storage_data["size"] = val.group(0)
        elif self.storage:
            self.storage["data"] = {"size": val.group(0)}
        else:
            self.spec["storage"] = {"data": {"size": val.group(0)}}

    @property
    def storage_logs(self):
        """
        Returns the log volume details of the SQL managed instance
        :param self:
        :return:
        """
        if self.storage:
            return self.storage.get("logs")
        else:
            return None

    @property
    def storage_class_logs(self):
        """
        Returns the storage class of the log volume
        :param self:
        :return:
        """
        if self.storage_logs:
            return self.storage_logs.get("className")
        else:
            return None

    @storage_class_logs.setter
    def storage_class_logs(self, sc):
        """
        Sets the storage class of the log volume.
        :param self:
        :param sc: storage class
        :return:
        """
        if self.storage_logs:
            self.storage_logs["className"] = sc
        elif self.storage:
            self.storage["logs"] = {"className": sc}
        else:
            self.spec["storage"] = {"logs": {"className": sc}}

    @property
    def storage_datalogs(self):
        """
        Returns the data log volume details of the SQL managed instance
        :param self:
        :return:
        """
        if self.storage:
            return self.storage.get("datalogs")
        else:
            return None
    @property
    def storage_class_datalogs(self):
        """
        Returns the storage class of the data logs volume
        :param self:
        :return:
        """
        if self.storage_datalogs:
            return self.storage_datalogs.get("className")
        else:
            return None
    @storage_class_datalogs.setter
    def storage_class_datalogs(self, sc):
        """
        Sets the storage class of the data logs volume.
        :param self:
        :param sc: storage class
        :return:
        """
        if self.storage_datalogs:
            self.storage_datalogs["className"] = sc
        elif self.storage:
            self.storage["datalogs"] = {"className": sc}
        else:
            self.spec["storage"] = {"datalogs": {"className": sc}}

    @property
    def volume_size_logs(self):
        """
        Returns the volume size of the logs volume
        :param self:
        :return:
        """
        if self.storage_logs:
            return self.storage_logs.get("size")
        else:
            return None

    @volume_size_logs.setter
    def volume_size_logs(self, vs):
        """
        Sets the volume size of the logs volume.
        :param self:
        :param sc: storage class
        :return:
        """
        val = re.search(r"^([1-9]+[0-9]*)(Gi|Mi|Ki)$", vs)
        if not val:
            raise ValueError("Volume size must be a positive integer ending in `Ki`, `Mi`, or `Gi`.")

        if self.storage_logs:
            self.storage_logs["size"] = val.group(0)
        elif self.storage:
            self.storage["logs"] = {"size": val.group(0)}
        else:
            self.spec["storage"] = {"logs": {"size": val.group(0)}}

    @property
    def volume_size_datalogs(self):
        """
        Returns the volume size of the data logs volume
        :param self:
        :return:
        """
        if self.storage_datalogs:
            return self.storage_datalogs.get("size")
        else:
            return None
    @volume_size_datalogs.setter
    def volume_size_datalogs(self, vs):
        """
        Sets the volume size of the data logs volume.
        :param self:
        :param sc: storage class
        :return:
        """
        val = re.search(r"^([1-9]+[0-9]*)(Gi|Mi|Ki)$", vs)
        if not val:
            raise ValueError("Volume size must be a positive integer ending in `Ki`, `Mi`, or `Gi`.")
        if self.storage_datalogs:
            self.storage_datalogs["size"] = val.group(0)
        elif self.storage:
            self.storage["datalogs"] = {"size": val.group(0)}
        else:
            self.spec["storage"] = {"datalogs": {"size": val.group(0)}}

    @property
    def storage_backups(self):
        """
        Returns the backups volume details of the SQL managed instance
        :param self:
        :return:
        """
        if self.storage:
            return self.storage.get("backups")
        else:
            return None

    @property
    def storage_class_backups(self):
        """
        Returns the storage class of the backup volume
        :param self:
        :return:
        """
        if self.storage_backups:
            return self.storage_backups.get("className")
        else:
            return None

    @storage_class_backups.setter
    def storage_class_backups(self, sc):
        """
        Sets the storage class of the backup volume.
        :param self:
        :param sc: storage class
        :return:
        """
        if self.storage_backups:
            self.storage_backups["className"] = sc
        elif self.storage:
            self.storage["backups"] = {"className": sc}
        else:
            self.spec["storage"] = {"backups": {"className": sc}}

    @property
    def volume_size_backups(self):
        """
        Returns the volume size of the backups volume
        :param self:
        :return:
        """
        if self.storage_backups:
            return self.storage_backups.get("size")
        else:
            return None

    @volume_size_backups.setter
    def volume_size_backups(self, vs):
        """
        Sets the volume size of the backups volume.
        :param self:
        :param sc: storage class
        :return:
        """
        val = re.search(r"^([1-9]+[0-9]*)(Gi|Mi|Ki)$", vs)
        if not val:
            raise ValueError("Volume size must be a positive integer ending in `Ki`, `Mi`, or `Gi`.")

        if self.storage_backups:
            self.storage_backups["size"] = val.group(0)
        elif self.storage:
            self.storage["backups"] = {"size": val.group(0)}
        else:
            self.spec["storage"] = {"backups": {"size": val.group(0)}}

    @property
    def scale(self):
        """
        Returns the scale property.
        :param self:
        :return:
        """
        return self.spec.get("scale")

    @scale.setter
    def scale(self, s):
        """
        Sets the scale property.
        :param self:
        :param s:
        :return:
        """
        self.spec["scale"] = s

    @property
    def replicas(self):
        """
        Returns the replicas property.
        :return:
        """
        if self.scale:
            return self.scale.get("replicas")
        else:
            return None

    @replicas.setter
    def replicas(self, rep):
        """
        Sets the replicas property.
        :param self:
        :param rep:
        :return:
        """
        try:
            val = int(rep)
            if val <= 0:
                raise ValueError()
        except ValueError:
            raise ValueError("Replicas must be a positive integer, with a minimum of 1.")
        if self.scale:
            self.scale["replicas"] = val
        else:
            self.spec["scale"] = {"replicas": val}

    @property
    def shards(self):
        """
        Returns the shards property.
        :return:
        """
        if self.scale:
            return self.scale.get("shards", 0)
        else:
            return 0

    @shards.setter
    def shards(self, s):
        """
        Sets the shards property.
        :param self:
        :param s:
        :return:
        """
        try:
            val = int(s)
            if val < 0:
                raise ValueError()
        except ValueError:
            raise ValueError("Shards must be a nonnegative integer.")
        if s:
            self.spec.setdefault('scale', {})['shards'] = val
        else:
            self.spec.get('scale', {}).pop('shards', None)

    @property
    def limits(self):
        """
        Returns the limits property.
        :param self:
        :return:
        """
        return self.spec.get("limits")

    @limits.setter
    def limits(self, lim):
        """
        Sets the limits property.
        :param self:
        :param lim:
        :return:
        """
        self.spec["limits"] = lim

    @property
    def memory_limit(self):
        """
        Returns the memory limit property.
        :return:
        """
        if self.limits:
            return self.limits.get("memory")
        else:
            return None

    @memory_limit.setter
    def memory_limit(self, lim):
        """
        Sets the memory limit property.
        :param self:
        :param lim:
        :return:
        """
        val = re.search(r"^([1-9]+[0-9]*)(Gi|Mi|Ki)$", lim)
        if not val:
            raise ValueError("Memory limit must be a positive integer ending in `Ki`, `Mi`, or `Gi`.")

        if self.limits:
            self.limits["memory"] = val.group(0)
        else:
            self.spec["limits"] = {"memory": val.group(0)}

    @property
    def cores_limit(self):
        """
        Returns the vcore limit property.
        :return:
        """
        if self.limits:
            return self.limits.get("vcores")
        else:
            return None

    @cores_limit.setter
    def cores_limit(self, lim):
        """
        Sets the vcore limit property.
        :param self:
        :param lim:
        :return:
        """
        try:
            val = int(lim)
            if val <= 0:
                raise ValueError()
        except ValueError:
            raise ValueError("Cores limit must be a positive integer, with a minimum of 1.")
        if self.limits:
            self.limits["vcores"] = lim
        else:
            self.spec["limits"] = {"vcores": lim}

    @property
    def requests(self):
        """
        Returns the requests property.
        :param self:
        :return:
        """
        return self.spec.get("requests")

    @requests.setter
    def requests(self, req):
        """
        Sets the requests property
        :param self:
        :param req:
        :return:
        """
        self.spec["requests"] = req

    @property
    def memory_request(self):
        """
        Returns the memory request property.
        :return:
        """
        if self.requests:
            return self.requests.get("memory")
        else:
            return None

    @memory_request.setter
    def memory_request(self, request):
        """
        Sets the memory request property.
        :param self:
        :param request:
        :return:
        """
        val = re.search(r"^([1-9]+[0-9]*)(Gi|Mi|Ki)$", request)
        if not val:
            raise ValueError("Memory request must be a positive integer ending in `Ki`, `Mi`, or `Gi`.")

        if self.requests:
            self.requests["memory"] = val.group(0)
        else:
            self.spec["requests"] = {"memory": val.group(0)}

    @property
    def cores_request(self):
        """
        Returns the cores request.
        :return:
        """
        if self.requests:
            return self.requests.get("vcores")
        else:
            return None

    @cores_request.setter
    def cores_request(self, request):
        """
        Sets the vcore request property.
        :param self:
        :param request:
        :return:
        """
        try:
            val = int(request)
            if val <= 0:
                raise ValueError()
        except ValueError:
            raise ValueError("Cores request must be a positive integer, with a minimum of 1.")
        if self.requests:
            self.requests["vcores"] = request
        else:
            self.spec["requests"] = {"vcores": request}

    def validate_memory(self):
        """
        To be called after fields are initialized - This method validates that the memory request is less than the
        memory limit and raises errors if not
        """
        request = self.memory_request
        limit = self.memory_limit
        if request and limit:
            request_size = MemoryResource(request)
            limit_size = MemoryResource(limit)
            if limit_size < request_size:
                raise ValueError("Memory request of %s cannot exceed memory limit of %s" % (request, limit))

    def validate_cores(self):
        """
        To be called after fields are initialized - This method validates that the cores request is less than the
        cores limit and raises errors if not
        """
        request = self.cores_request
        limit = self.cores_limit
        if request and limit:
            cores_limit = CpuResource(limit)
            cores_request = CpuResource(request)
            if cores_limit < cores_request:
                raise ValueError("Cores request of %s cannot exceed cores limit of %s" % (request, limit))
