import re
from abc import (ABC, abstractmethod)
from azdata.cli.core.constants import (
    ResourceType,
    MemoryFormat,
    CpuFormat,
)
from azdata.cli.core.util import (
    convert_cpu_size_to_millicores,
    convert_mem_size_to_bytes
)


class KubeResource(ABC):
    """
    An abstract kubernetes resource such as Memory/Cores/Pages. Includes parsing/validation/conversion to standard units.
    Also overrides comparison methods to work with standard python binary operators.
    """
    _ALL_QUANTITY_REGEX = r"^+?((([0-9]+)|([0-9]*\.[0-9]+))([MGTPE]i?|Ki|k|e[1-9]+))|([1-9]+[0-9]*|[0-9]*\.[0-9]+m?)$"
    VALID_QUANTITY_MSG = "Valid kubernetes resource quantities match {}."\
                         " See kubernetes documentation for more details".format(_ALL_QUANTITY_REGEX)

    def __init__(self, resource_type: ResourceType, quantity: str):
        self._resource_type = resource_type
        self.quantity = quantity

    @property
    def type(self):
        return self._resource_type

    @property
    def quantity(self):
        """
        The quantity of this resource in string format e.g. 10Gi, .6Mi, 10k, 1m
        """
        return self._quantity

    @quantity.setter
    def quantity(self, q):
        if hasattr(self, "_quantity"):
            raise AttributeError("Resource quantities are immutable")

        if re.compile(self.regex()).match(q) is None:
            raise ValueError(
                "Invalid kubernetes resource format {}. {}".format(q, self.VALID_QUANTITY_MSG))
        self._quantity = q

    @staticmethod
    def regex():
        """
        The regex that matches any quantity in this class of resource
        :return: regex that matches quantities of this type of resource
        """
        return KubeResource._ALL_QUANTITY_REGEX

    @staticmethod
    @abstractmethod
    def size():
        """
        Gives a numeric size of this resource that is comparable to other resources of the same type e.g. a memory
        resource will return size in bytes and a Cpu resource will return size in millicores
        :return: numeric size of this resource in some standard unit
        """
        raise NotImplementedError(
            "Abstract class KubeResource has no sane way to compute size")

    @staticmethod
    def is_valid_format(fmt):
        return re.compile(KubeResource.regex()).match(fmt) is not None

    def __str__(self):
        """
        @override
        """
        return self.quantity

    def __repr__(self):
        """
        @override
        """
        return "{}({})".format(type(self), self.quantity)

    def __eq__(self, other):
        """
        @override
        """
        if type(self) is not type(other):
            raise NotImplementedError("Cannot compare different resource types")
        return self.size == other.size

    def __ne__(self, other):
        """
        @override
        """
        if type(self) is not type(other):
            raise NotImplementedError("Cannot compare different resource types")
        return not (self == other)

    def __lt__(self, other):
        """
        @override
        """
        if type(self) is not type(other):
            raise NotImplementedError("Cannot compare different resource types")
        return self.size < other.size

    def __gt__(self, other):
        """
        @override
        """
        if type(self) is not type(other):
            raise NotImplementedError("Cannot compare different resource types")
        return self.size > other.size

    def __le__(self, other):
        """
        @override
        """
        if type(self) is not type(other):
            raise NotImplementedError("Cannot compare different resource types")
        return self.size <= other.size

    def __ge__(self, other):
        """
        @override
        """
        if type(self) is not type(other):
            raise NotImplementedError("Cannot compare different resource types")
        return self.size >= other.size


class MemoryResource(KubeResource):
    """
    Represents a kubernetes memory resource
    """
    VALID_QUANTITY_MSG = "Valid memory formats match {}."\
        "See kubernetes documentation for more details.".format(MemoryFormat.ALL.value)

    def __init__(self, quantity: str):
        super().__init__(ResourceType.MEMORY, quantity)
        self._bytes = None

    @staticmethod
    def regex():
        """
        @override
        """
        return MemoryFormat.ALL.value

    @property
    def size(self):
        """
        @override
        """
        if self._bytes is None:
            self._bytes = convert_mem_size_to_bytes(self.quantity)

        return self._bytes


class CpuResource(KubeResource):
    """
    Represents a kubernetes CpuResource
    """
    VALID_QUANTITY_MSG = "Valid Cpu resource quantities are strictly positive and match {}."\
                         " See kubernetes documentation for more details.".format(CpuFormat.ALL.value)

    def __init__(self, quantity):
        super().__init__(ResourceType.CPU, quantity)
        self._millicores = None

    @staticmethod
    def regex():
        """@override"""
        return CpuFormat.ALL.value

    @property
    def size(self):
        """
        @override
        """
        if self._millicores is None:
            self._millicores = convert_cpu_size_to_millicores(self.quantity)

        return self._millicores
